#!/usr/bin/perl -w

# Copyright 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use warnings;
use File::Spec;
use FLAT;
use List::Util 'max','min','sum','uniqstr';
use Math::BaseCnv 'cnv';
use Math::BigInt try => 'GMP';
use Math::PlanePath::GrayCode;
use Math::PlanePath::Base::Digits
  'bit_split_lowtohigh','digit_join_lowtohigh';
use POSIX 'ceil';

use FindBin;
use lib::abs "$FindBin::Bin/../tools";
use MyFLAT;
use MyATT;
$|=1;

# uncomment this to run the ### lines
# use Smart::Comments;

my $Udelta = 'ab w bc x ca y ba z ac';
my @Udelta = split / */, $Udelta;
print @Udelta,"\n";
my %Usymbol_count;
foreach my $s (@Udelta) {
  $Usymbol_count{$s}++;
}
my $Udelta_len = scalar(@Udelta);
my $Udelta_num_symbols = scalar keys %Usymbol_count;
print "Udelta ",@Udelta,"  length $Udelta_len distinct $Udelta_num_symbols\n";


{
  # Udelta breadth first by prefixing
  #
  # counts cf
  # A006156 ternary square-free words
  # A051042 ternary cube-free words
  #
  # n=15
  # not in OEIS: 1594323, 4780782, 14313186, 42703536, 126628962
  # digits(42703536,3)
  # 3^14 - 4780782 == 3^7
  # 3^15 - 14313186 == 35721
  # 3^15 - 14313186 == 3^6*(3^4 + 2*3^2 - 1)/2
  # digits((3^10 + 2*3^8 - 3^6)/2 ,3)
  # digits(3^6*(3^4 + 2*3^2 - 1)/2 ,3)
  # 3^16 - 42703536 == 343185
  # 3^17 - 126628962 == 2511201
  # digits(2511201,3) == [1, 1, 2, 0, 1, 1, 2, 0, 2, 0, 1, 1, 1, 0]
  # #digits(2511201,3) == 14


  # Binary:
  # len=0  in 1 out 2  not 0  outsize 4 /z/tmp/b001.txt
  # len=1  in 2 out 4  not 0  outsize 12 /z/tmp/b002.txt
  # len=2  in 4 out 8  not 0  outsize 32 /z/tmp/b003.txt
  # len=3  in 8 out 16  not 0  outsize 80 /z/tmp/b004.txt
  # len=4  in 16 out 32  not 0  outsize 192 /z/tmp/b005.txt
  # len=5  in 32 out 64  not 0  outsize 448 /z/tmp/b006.txt
  # len=6  in 64 out 128  not 0  outsize 1024 /z/tmp/b007.txt
  # len=7  in 128 out 256  not 0  outsize 2304 /z/tmp/b008.txt
  # len=8  in 256 out 512  not 0  outsize 5120 /z/tmp/b009.txt
  # len=9  in 512 out 1024  not 0  outsize 11264 /z/tmp/b010.txt
  # len=10  in 1024 out 2048  not 0  outsize 24576 /z/tmp/b011.txt
  # len=11  in 2048 out 4096  not 0  outsize 53248 /z/tmp/b012.txt
  # len=12  in 4096 out 8192  not 0  outsize 114688 /z/tmp/b013.txt
  # len=13  in 8192 out 16256  not 128  outsize 243840 /z/tmp/b014.txt
  # len=14  in 16256 out 31552  not 960  outsize 504832 /z/tmp/b015.txt
  # len=15  in 31552 out 58656  not 4448  outsize 997152 /z/tmp/b016.txt
  # len=16  in 58656 out 102090  not 15222  outsize 1837620 /z/tmp/b017.txt
  # len=17  in 102090 out 162830  not 41350  outsize 3093770 /z/tmp/b018.txt
  # len=18  in 162830 out 234352  not 91308  outsize 4687040 /z/tmp/b019.txt
  # len=19  in 234352 out 302588  not 166116  outsize 6354348 /z/tmp/b020.txt
  # len=20  in 302588 out 352618  not 252558  outsize 7757596 /z/tmp/b021.txt
  # len=21  in 352618 out 376950  not 328286  outsize 8669850 /z/tmp/b022.txt
  # len=22  in 376950 out 377150  not 376750  outsize 9051600 /z/tmp/b023.txt
  # len=23  in 377150 out 358168  not 396132  outsize 8954200 /z/tmp/b024.txt
  # len=24  in 358168 out 323878  not 392458  outsize 8420828 /z/tmp/b025.txt
  # len=25  in 323878 out 277598  not 370158  outsize 7495146 /z/tmp/b026.txt
  # len=26  in 277598 out 224324  not 330872  outsize 6281072 /z/tmp/b027.txt
  # len=27  in 224324 out 170170  not 278478  outsize 4934930 /z/tmp/b028.txt
  # len=28  in 170170 out 121286  not 219054  outsize 3638580 /z/tmp/b029.txt
  # len=29  in 121286 out 81462  not 161110  outsize 2525322 /z/tmp/b030.txt
  # len=30  in 81462 out 52296  not 110628  outsize 1673472 /z/tmp/b031.txt

  # my(len=13); len+1-(7-1) == 8
  # 8/4 == 2

  my $radix = 3;
  my $verbose = 0;

  print "Udelta $Udelta\n";
  my $t = time();

  # print "original  $Udelta\n";
  # my $T = $Udelta;
  # $Udelta = reverse $Udelta;
  # print "reversed  $Udelta\n";
  # $T =~ tr/abc wxyz/cab zyxw/;
  # print "transform $T\n";
  # # $Udelta = 'xxx';

  my $tempdir = File::Spec->catdir(File::Spec->tmpdir,
                                   'Udelta');
  print "temporary directory $tempdir\n";
  unless (-d $tempdir) {
    mkdir $tempdir or die "oops, cannot make $tempdir: $!";
  }
  my $len_to_filename = sub {
    my ($len) = @_;
    return File::Spec->catfile($tempdir,
                               sprintf('radix%d-%03d.txt', $radix, $len));
  };
  open my $fh, '>', $len_to_filename->(0) or die;
  print $fh "\n";
  close $fh or die;

  my @avoid_count;
  foreach my $len (0 .. 999) {
    my %Usymbol_to_re;
    my $re = '';
    my $capnum = 0;
    foreach my $Usymbol (@Udelta) {
      if ($Usymbol_to_re{$Usymbol}) {
        $re .= $Usymbol_to_re{$Usymbol};
        next;
      }
      # $len+1 new string length
      # $Udelta_len subtract own count is how much of Udelta len taken by
      # at least 1 for each other symbol
      # Divide by own count for maximum length of this.
      #
      my $max = int(($len+1 - ($Udelta_len-$Usymbol_count{$Usymbol}))
                    / $Usymbol_count{$Usymbol});
      if ($max < 1) {
        if ($verbose) {
          print "no match possible for len=$len symbol $Usymbol\n";
        }
        $max = 1;
      }
      my $term = ($max <= 1 ? "." : ".{1,$max}");
      if ($Usymbol_count{$Usymbol} <= 1) {
        $re .= $term;
      } else {
        $re .= "($term)";
        $capnum++;
        $Usymbol_to_re{$Usymbol} = "\\$capnum";
      }
    }
    $|=1;
    if ($verbose) {
      print "regexp $re\n";
    }
    $re = qr/^$re/o;     # anchored as new symbol is at start


    $|=0;
    my $in_count = 0;
    my $out_count = 0;
    my $not_count = 0;
    my $in_filename = $len_to_filename->($len);
    open my $fh, '<', $in_filename or die "oops, cannot open $in_filename: $!";
    my $out_filename = $len_to_filename->($len+1);
    open my $out, '>', $out_filename or die;
    while (defined(my $str = readline $fh)) {
      chomp $str;
      $in_count++;
      foreach my $s (0 .. $radix-1) {
        my $new = $s . $str;
        if ($new =~ $re) {  # encounters Udelta
          $not_count++;
        } else {            # avoids Udelta
          print $out $new,"\n" or die "oops, error writing: $!";
          $out_count++;
        }
      }
    }
    close $out or die "oops, error writing: $!";
    $avoid_count[$len] = $in_count;
    my $out_size = -s $out_filename;
    $|=1;
    if ($verbose) {
      print "len=$len  in $in_count not $not_count out $out_count outfile $out_filename size $out_size\n";
    } else {
      print "$out_count,";
    }
    last unless $in_count;
  }

  my $total = 0;
  foreach my $len (0 .. $#avoid_count) {
    print $avoid_count[$len],',';
    if ($len==18 || $len==30) { print "\n"; }
    $total += $avoid_count[$len];
  }
  print "\ntotal $total\n";

  $t = (time() - $t)/60;
  printf "runtime %.1f minutes\n", $t;
  exit 0;
}
{
  # 2-avoiding counts for %S,%T,%U
  my $len_to_filename = sub {
    my ($len) = @_;
    return sprintf '/z/tmp/b%03d.txt', $len;
  };
  open my $fh, '>', $len_to_filename->(0) or die;
  print $fh "\n";
  close $fh or die;

  my $total = 0;
  foreach my $len (0 .. 51) {
    my $count = 0;
    open my $fh, '<', $len_to_filename->($len) or die;
    while (defined(my $str = readline $fh)) {
      $count++;
    }
    print "$count,";
    if ($len==18 || $len==30) { print "\n"; }
    $total += $count;
  }
  print "\ntotal $total\n";
  exit 0;
}

{
  # Foma of particular length
  # 17-14 == 3
  # 35721

  require IPC::Run;

  my $Udelta = 'ab w bc x ca y ba z ac';
  $Udelta =~ s/ //g;
  my @U = split //, $Udelta;

  my @strings_by_length = (undef);
  foreach my $len (1 .. 3) {
    foreach my $n (0 .. 3**$len-1) {
      my $str = sprintf '%0*s', $len, cnv($n,10,3);
      $str =~ tr/012/123/;
      $str =~ s/./ $&/g;
      push @{$strings_by_length[$len]}, $str;
    }
  }
  ### @strings_by_length

  my $len = 18;

  open my $fh, '>', '/tmp/Udelta.foma' or die;
  my $stack = 0;
  # print $fh "regex 0;\n";
  # $stack++;
  my $any = ' [1|2|3]';

 A: for (my $alen = 1; $len - 4*$alen >= 6; $alen++) {
    my $rlen = $len - 4*$alen;
    foreach my $a (@{$strings_by_length[$alen] // die}) {

      for (my $blen = 1; $rlen - 3*$blen >= 5; $blen++) {
        my $rlen = $rlen - 3*$blen;
        foreach my $b (@{$strings_by_length[$blen] // die}) {

          for (my $clen = 1; $rlen - 3*$clen >= 4; $clen++) {
            my $rlen = $rlen - 3*$clen;
            foreach my $c (@{$strings_by_length[$clen] // die}) {

              for (my $wlen = 1; $rlen- $wlen >= 3; $wlen++) {
                my $rlen = $rlen - $wlen;
                my $w = $any x $wlen;

                for (my $xlen = 1; $rlen- $xlen >= 2; $xlen++) {
                  my $rlen = $rlen - $xlen;
                  my $x = $any x $xlen;

                  for (my $ylen = 1; $rlen- $ylen >= 1; $ylen++) {
                    my $rlen = $rlen - $ylen;
                    my $y = $any x $ylen;

                    for (my $zlen = 1; $rlen- $zlen >= 0; $zlen++) {
                      my $rlen = $rlen - $zlen;
                      my $z = $any x $zlen;

                      for (my $prelen = 0; $rlen-$prelen >= 0; $prelen++) {
                        my $postlen = $rlen - $prelen;
                        my $pre = $any x $prelen;
                        my $post = $any x $postlen;

                        my @lens = ($prelen,
                                    $alen,$blen,  $wlen,  $blen,$clen,
                                    $xlen, $clen,$alen, $ylen, $blen,$alen,
                                    $zlen, $alen,$clen,
                                    $postlen);
                        sum(@lens) == $len or die;
                        # print "lens: @lens\n";

                        # next if $prelen || $postlen;

                        my $re = "$pre $a$b $w $b$c $x $c$a $y $b$a $z $a$c $post";
                        # $re = '1 1';
                        # $re = '~[[2|3] [1|2|3]*]';
                        # $re = "[[1|2|3]*] & ~[$re]";
                        # print $fh "echo a=$a b=$b c=$c\n";
                        # print $fh "print net\n";
                        # print $fh "print net\n";
                        # print $fh "print upper-words\n";
                        # print $fh "test null\n";
                        # last A;

                        print $fh "regex $re;\n";
                        $stack++;
                        if ($stack >= 2) {
                          print $fh "union net\n";
                          $stack--;
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  print $fh "write att /tmp/Udelta.att\n";
  # print $fh "print upper-words\n";
  close $fh or die;

  system 'foma -f /tmp/Udelta.foma | tail -20';

  exit;
  my $att = MyATT->read_file('/tmp/Udelta.att');
  my $f = $att->as_FLAT;
  print join(',',MyFLAT::FLAT_count_contains($f,Math::BigInt->new($len+2))),"\n";
  # $f->MyFLAT::view();

  # my $h = IPC::Run::harness(['echo'], '<pipe', \*FH);
  # $h->start;
  # $h->pump;
  # $h->finish or die "$?";

  # my $str = "regex $start;\n";
  # print $str;
  # FOMA_run($str, no_capture => 1);
  exit 0;
}
{
  # Foma excludes
  # becomes too big

  require IPC::Run;

  my $Udelta = 'ab w bc x ca y ba z ac';
  $Udelta =~ s/ //g;
  my @U = split //, $Udelta;

  my $start = '(1+ (2 [1|2|3]*))';
  #$start = '[1|2|3]*';

  open my $fh, '|foma' or die;
  print $fh "regex $start;\n";
  print $fh "print net\n";
  print $fh "print upper-words\n";

  my @strings = ('1','2','3');
 A: foreach my $a (@strings) {
    foreach my $b (@strings) {
      foreach my $c (@strings) {
        my $re = "~[ [1|2|3]* $a $b [1|2|3]* $b $c [1|2|3]* $c $a [1|2|3]* $b $a [1|2|3]* $a $c [1|2|3]* ]";
        # $re = '1 1';
        # $re = '~[[2|3] [1|2|3]*]';
        # $re = "[[1|2|3]*] & ~[$re]";
        print $fh "echo a=$a b=$b c=$c\n";
        print $fh "regex $re;\n";
        # print $fh "print net\n";
        print $fh "intersect net\n";
        # print $fh "print net\n";
        # print $fh "print upper-words\n";
        # print $fh "test null\n";
        # last A;
      }
    }
  }
  # my $h = IPC::Run::harness(['echo'], '<pipe', \*FH);
  # $h->start;
  # $h->pump;
  # $h->finish or die "$?";

  # my $str = "regex $start;\n";
  # print $str;
  # FOMA_run($str, no_capture => 1);
  exit 0;
}
{
  # Udelta, AG(Udelta), Graph
  # https://hog.grinvin.org/ViewGraphInfo.action?id=34017

  require Graph;
  require MyGraphs;

  my $Udelta = 'ab w bc x ca y ba z ac';
  $Udelta =~ s/ //g;

  my @alphabet = sort {$a cmp $b} uniqstr(split //, $Udelta);
  print "$Udelta\n";
  print "alphabet ",@alphabet,"\n";

  my $graph = Graph->new (undirected => 1);
  foreach my $i (0 .. $#alphabet) {
    my $r = $alphabet[$i].'.R';
    my $l = $alphabet[$i].'.L';
    $graph->add_vertices($l,$r);
    MyGraphs::Graph_set_xy_points($graph, $r => [0,-$i], $l => [4,-$i]);
  }
  my @U = split //, $Udelta;
  foreach my $i (0 .. $#U-1) {
    $graph->add_edge($Udelta[$i].'.R', $Udelta[$i+1].'.L');
  }

  foreach my $v ($graph->vertices) {
    if ($graph->degree($v) >= 4) {
      $v eq 'a.R' or die;
    }
  }

  MyGraphs::hog_searches_html($graph);
  MyGraphs::Graph_view($graph);

  MyGraphs::Graph_set_xy_points
      ($graph,
       'y.R' => [-1,1],
       'b.L' => [-1,0], 'w.R' => [-1,-1],
       'a.R' => [0,0], 'y.L' => [0,1], 'z.L' => [0,-1],
       'c.L' => [1,0], 'x.R' => [1,-1],
       'b.R' => [2,0], 'w.L' => [2,-1],
       'a.L' => [3,0], 'z.R' => [3,-1],
       'c.R' => [4,0], 'x.L' => [4,-1],
      );
  MyGraphs::Graph_view($graph);
  MyGraphs::hog_upload_html($graph);
  exit 0;
}



{
  # regex by prefixing

  # len=143  22222222222221212121212121212121212121202111111111111102010201020102010201020102010201020102010201020102010101010101010101010101010000000000000  tries=8718

  my $Udelta = 'ab w bc x ca y ba z ac';
  $Udelta =~ s/ //g;
  my @U = split //, $Udelta;

  my @w = (((0)x13), ((1,0)x13), ((2,0,1,0)x13));
  @w == 91 or die;
  my $w = join(@w);

  # @U = ('a','a');
  # $w = 'abcdbcdba';
  # $w = 'abacabcacbabcabacabc';

  my %Usymbol_to_re;
  my $re = '';
  foreach my $s (@U) {
    if ($Usymbol_to_re{$s}) {
      $re .= $Usymbol_to_re{$s};
    } else {
      $re .= '(.+)';
      $Usymbol_to_re{$s} = '\\'.(scalar(keys %Usymbol_to_re)+1);
    }
  }
  print "$re\n";
  $re = qr/^$re/o;
  my $match = ($w =~ $re ? "yes" : "no");
  print "$match\n";

  no warnings 'recursion';
  my $tries = 0;
  my $max;
  my $try;
  $try = sub {
    my ($str_ref) = @_;
    if (length($$str_ref) > length($max)) {
      print "len=",length($$str_ref),"  $$str_ref  tries=$tries\n";
      $max = $$str_ref;
      $tries = 0;
    }
    foreach my $s (0 .. min(2,length($$str_ref))) {
      print "len=",length($$str_ref)," $s\r";
      my $new = $s . $$str_ref;
      if ($new =~ $re) {
        $tries++;
      } else {
        $try->(\$new);
      }
    }
  };
  $try->(\'0');
  exit;
}
{
  # len=143  00000000000001010101010101010101010101020102010201020102010201020102010201020102010201020111111111111121212121212121212121212122022222022202012  tries=66

  my $Udelta = 'ab w bc x ca y ba z ac';
  $Udelta =~ s/ //g;
  my @U = split //, $Udelta;

  my @w = (((0)x13), ((1,0)x13), ((2,0,1,0)x13));
  @w == 91 or die;
  my $w = join(@w);

  # @U = ('a','a');
  # $w = 'abcdbcdba';
  # $w = 'abacabcacbabcabacabc';

  my %Usymbol_to_re;
  my $re = '';
  foreach my $s (@U) {
    if ($Usymbol_to_re{$s}) {
      $re .= $Usymbol_to_re{$s};
    } else {
      $re .= '(.+)';
      $Usymbol_to_re{$s} = '\\'.(scalar(keys %Usymbol_to_re)+1);
    }
  }
  print "$re\n";
  $re = qr/$re/o;
  my $match = ($w =~ $re ? "yes" : "no");
  print "$match\n";

  no warnings 'recursion';
  my $tries = 0;
  my $max;
  my $try;
  $try = sub {
    my ($str_ref) = @_;
    if (length($$str_ref) > length($max)) {
      print "len=",length($$str_ref),"  $$str_ref  tries=$tries\n";
      $max = $$str_ref;
      $tries = 0;
    }
    foreach my $s (0 .. min(2,length($$str_ref))) {
      my $new = $$str_ref . $s;
      if ($new =~ $re) {
        $tries++;
      } else {
        $try->(\$new);
      }
    }
  };
  $try->(\'0');
  exit;
}

{
  # Foma excludes
  # becomes too big

  require IPC::Run;

  my $Udelta = 'ab w bc x ca y ba z ac';
  $Udelta =~ s/ //g;
  my @U = split //, $Udelta;

  my $start = '(1+ (2 [1|2|3]*))';
  #$start = '[1|2|3]*';

  open my $fh, '|foma' or die;
  print $fh "regex $start;\n";
  print $fh "print net\n";
  print $fh "print upper-words\n";

  my @strings = ('1','2','3');
 A: foreach my $a (@strings) {
    foreach my $b (@strings) {
      foreach my $c (@strings) {
        my $re = "~[ [1|2|3]* $a $b [1|2|3]* $b $c [1|2|3]* $c $a [1|2|3]* $b $a [1|2|3]* $a $c [1|2|3]* ]";
        # $re = '1 1';
        # $re = '~[[2|3] [1|2|3]*]';
        # $re = "[[1|2|3]*] & ~[$re]";
        print $fh "echo a=$a b=$b c=$c\n";
        print $fh "regex $re;\n";
        # print $fh "print net\n";
        print $fh "intersect net\n";
        # print $fh "print net\n";
        # print $fh "print upper-words\n";
        # print $fh "test null\n";
        # last A;
      }
    }
  }
  # my $h = IPC::Run::harness(['echo'], '<pipe', \*FH);
  # $h->start;
  # $h->pump;
  # $h->finish or die "$?";

  # my $str = "regex $start;\n";
  # print $str;
  # FOMA_run($str, no_capture => 1);
  exit 0;
}



