#!/usr/bin/perl -w

# Copyright 2014, 2015, 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


# Usage: perl centroid-on-boundary.pl
#
# Create some files
#
#     /tmp/centroid-001.png
#
# etc which are images of where the segments centroid is relative to the
# dragon curve.  View them with for example
#
#     xzgv /tmp/centroid-*.png
#
# The square containing the centroid is in the middle of each image.  This
# is the JEQ end square and develops in a 6-period pattern per the proofs.
#
# $want_rotate option below does a rotation by 45*$k so that the curve start
# to end is horizontal.  The way the square develops is less clear that way.


use 5.010;
use strict;
use Math::PlanePath::DragonCurve 118; # v.118 for xyxy_to_n_either()
use Geometry::AffineTransform;
use Image::Base::GD;

use FindBin;
use lib::abs "$FindBin::Bin/../tools";
use MyDragon ':all';


my $want_rotate = 0;

my $square_scale = 12;  # squares across picture
my $pixel_scale = 35;   # pixels per square
my $circle_size = .15;  # fraction of a square
my $dot_size = .1;      # fraction of a square

sub path_xyxy_is_traversed_either {
  my ($path, $x1,$y1, $x2,$y2, $n_limit) = @_;
  my $n = $path->xyxy_to_n_either($x1,$y1, $x2,$y2);
  return defined $n && $n < $n_limit;
}

my $path = Math::PlanePath::DragonCurve->new;
my @values;
foreach my $k (2 .. 100) {
  my ($n_lo, $n_hi) = $path->level_to_n_range($k);

  my ($x,$y) = GS_xy($k);
  my $x1 = int($x);                          # x1,y1,x2,y2 square
  my $x2 = ($x < $x1 ? $x1 - 1 : $x1 + 1);   # containing centroid
  my $y1 = int($y);
  my $y2 = ($y < $y1 ? $y1 - 1 : $y1 + 1);

  if ($x1 > $x2) { ($x1,$x2) = ($x2,$x1) }
  if ($y1 > $y2) { ($y1,$y2) = ($y2,$y1) }

  my $enc = (path_xyxy_is_traversed_either($path, $x1,$y1, $x2,$y1, $n_hi)
             && path_xyxy_is_traversed_either($path, $x1,$y2, $x2,$y2, $n_hi)
             && path_xyxy_is_traversed_either($path, $x1,$y1, $x1,$y2, $n_hi)
             && path_xyxy_is_traversed_either($path, $x2,$y1, $x2,$y2, $n_hi)
             ? 1 : 0);
  my $any = (path_xyxy_is_traversed_either($path, $x1,$y1, $x2,$y1, $n_hi)
             || path_xyxy_is_traversed_either($path, $x1,$y2, $x2,$y2, $n_hi)
             || path_xyxy_is_traversed_either($path, $x1,$y1, $x1,$y2, $n_hi)
             || path_xyxy_is_traversed_either($path, $x2,$y1, $x2,$y2, $n_hi)
             ? 1 : 0);
  print "k=$k  $enc,$any";
  my $xs = $x->numify;
  my $ys = $y->numify;
  if ($enc) {
    print "  $x $y  $xs $ys    $x1,$y1  $x2,$y2";
  }
  print "\n";
  push @values, $enc;

  if ($k > 0 && $k < 50) {
    my $affine = Geometry::AffineTransform->new;
    my $xm = ($x1->numify + $x2->numify)/2;  # middle of centroid square
    my $ym = ($y1->numify + $y2->numify)/2;

    $affine->translate(-$xm, -$ym);
    if ($want_rotate) {
      $affine->rotate(-45*$k);
    }
    $affine->translate($square_scale/2, - $square_scale/2);

    $affine->scale($pixel_scale,-$pixel_scale);

    my $width = $square_scale*$pixel_scale;
    my $height = $square_scale*$pixel_scale;
    my $image = Image::Base::GD->new(-width => $width,
                                     -height => $height);
    $image->rectangle (0,0, $width-1,$height-1, 'black', 1);
    {
      my ($wx,$wy) = $affine->transform($x->numify,$y->numify);
      $image->ellipse ($wx - $pixel_scale*$circle_size,
                       $wy - $pixel_scale*$circle_size,
                       $wx + $pixel_scale*$circle_size,
                       $wy + $pixel_scale*$circle_size,
                       'red');
    }
    foreach my $x ($x1 - 6 .. $x2 + 6) {
      foreach my $y ($y1 - 6 .. $y2 + 6) {
        foreach my $vert (0,1) {
          my $tx = $x + ($vert ? 0 : 1);
          my $ty = $y + ($vert ? 1 : 0);
          my $n = $path->xyxy_to_n_either($x,$y, $tx,$ty) // next;
          $n < $n_hi or next;

          my $colour = ($n % 2 ? 'blue' : 'white');
          $colour = 'white';
          $colour = ($n < $n_hi/2 ? 'white' : 'blue');
          $image->line($affine->transform($x,$y),
                       $affine->transform($tx,$ty),
                       $colour);

          if (($x+$y) % 2) {
            my ($wx,$wy) = $affine->transform($x,$y);
            $image->ellipse ($wx - $pixel_scale*$dot_size,
                             $wy - $pixel_scale*$dot_size,
                             $wx + $pixel_scale*$dot_size,
                             $wy + $pixel_scale*$dot_size,
                             'green', 1);
          }
        }
      }
    }
    $image->rectangle (0,0, 10,10, $enc ? 'green' : 'red', 1);
    $image->save(sprintf "/tmp/centroid-%02d.png", $k);
  }
}

require Math::OEIS::Grep;
Math::OEIS::Grep->search (array => \@values,verbose=>1);
exit 0;
