#!/usr/bin/perl -w

# Copyright 2019, 2020 Kevin Ryde
#
# ispell-batch-tex.pl is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation; either version 3, or (at your option) any
# later version.
#
# ispell-batch-tex.pl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with ispell-batch-tex.pl.  If not, see <http://www.gnu.org/licenses/>.


# Usage: ispell-batch-tex.pl filename.tex...
#
# See POD below.
#

use 5.010;
use strict;
use warnings;
use FindBin;
use Regexp::Common 'balanced';
use File::Slurp;
use Getopt::Long;
use Lingua::Ispell;
use List::Util 'max';

our $VERSION = 0;
my $option_verbose = 0;
my $exit_code = 0;
$|=1;

# Lingua::Ispell version 0.07 has a bug where it sends '-' instead of '+'
# for the named mode.  Might prefer to crunch out TeX macros here anyway.
#
# Lingua::Ispell::parse_according_to('tex');
Lingua::Ispell::_send_command('+', 'tex');

# Can run ispell in TeX mode from command line option too.
# $Lingua::Ispell::options{'-t'} = [];
# $Lingua::Ispell::options{'-T'} = ['tex'];
# $Lingua::Ispell::options{'-k'} = ['texskip1'];
#
# Are texskips extentable ?
# Lingua::Ispell::_send_command('-k', 'texskip1');
# Lingua::Ispell::_send_command("\-", 'tex texskip1');

# https://rt.cpan.org/Public/Dist/Display.html?Name=Lingua-Ispell

# TeX Book chapter 24 page 270 dimensions syntax spec
my $internal_unit_re = qr/e[mx]/;
my $physical_unit_re = qr/bp|c[cm]|dd|p[ct]|in|m[mu]|sp/;
my $unit_re = qr/$internal_unit_re|(true )?($physical_unit_re)/;
my $number_re = qr/[+-]?[ \t]*[0-9]*([0-9]|[.,])[0-9]*/;
my $dimension_re = qr/$number_re[ \t]*$unit_re/o;

sub spaces {
  my ($str) = @_;
  $str =~ tr/\n//cd;
  return $str;
}
CHECK {
  spaces("xx\n y\n") eq "\n\n" or die;
}

sub pos_linenum {
  my ($str, $pos) = @_;
  $pos //= pos($_[0]);
  $str = substr($str, 0, $pos);
  return 1 + scalar($str =~ tr/\n//);
}

sub tex_strip_comments {
  my ($str) = @_;
  # even number of backslashes  \\\\% ...
  # delete the % onwards, keep the backslashes
  $str =~ s/((^|[^\\])(\\\\)*)%[^\n]*/$1/mg;
  return $str;
}

sub check_file {
  my ($filename) = @_;
  if ($option_verbose >= 2) {
    print "$filename\n";
  }
  my $str = File::Slurp::read_file($filename, err_mode=>'quiet');
  if (! defined $str) {
    print STDERR "Cannot read $filename: $!\n";
    $exit_code = 1;
    return;
  }

  {
    # within comments
    my $localwords = 0;
    while ($str =~ /LocalWords:[ \t]*(.*)/mg) {
      foreach my $word (split /\s+/, $1) {
        if ($option_verbose >= 2) {
          print "LocalWords \"$word\"\n";
        }
        Lingua::Ispell::accept_word($word);
        $localwords++;
      }
    }
    if ($option_verbose >= 1) {
      print "$filename $localwords LocalWords\n";
    }
  }

  $str = tex_strip_comments($str);

  # \\ escaped backslashes, so that one backslash \foo is enough to identify
  # macros below
  $str =~ s/\\\\/  /g;

  # $str =~ s<https?://[^[:space:]}]*><spaces($&)>ge;
  $str =~ s<\\[-`'"~]><>g;           # accents, and optional hyphen
  $str =~ s<\\[Hcv]\{(.*?)\}><$1>g;  # accents \c{x} \v{x} \H{x}
  $str =~ s<\\([lL])\{\}><$1>g;      # stroke \l{} or \L{}
  $str =~ s<\bCo\.><Company.>g;      # abbreviation Co.

  # Don't think ispell pays much attention to hypens for say "co-linear"
  $str =~ s<\\hyp\{\}><->g;       # breakable hypen \hyp from hyphenat

  my $braces_re   = $RE{balanced}{-parens=>'{}'};
  my $square_brackets_re = $RE{balanced}{-parens=>'[]'};

  # a few dimensions commonly set
  $str =~ s{\\(tabcolsep)=$dimension_re}{' ' x length($&)}ge;

  # macros of 1-arg
  my $macro1_re = qr/(auto(page)?ref\*?|label|cite|bibitem
                     |foreignlanguage|addcontentsline
                     |usepackage|hypersetup|indexsetup
                     |usetikzlibrary|tikzset|tikzmath
                     |tocloftpagestyle
                     |newtagform|usetagform  # amsmath
                     |href|url               # hyperref
                     )/x;

  # macros of 2-args
  my $macro2_re = qr/(PassOptionsToPackage)/x;

  # macros of 3-args
  my $macro3_re = qr/(newfloat)/;

  # \makeindex[args...]
  $str =~ s<(\\[a-zA-Z]+)($square_brackets_re)><$1.spaces($2)>ge;

  # \usepackage[args...]{name}
  $str =~ s<\\$macro1_re($square_brackets_re)?$braces_re><spaces($&)>ge;
  $str =~ s<\\$macro2_re$braces_re$braces_re><spaces($&)>ge;
  $str =~ s<\\$macro3_re$braces_re$braces_re$braces_re><spaces($&)>ge;

  # print $Lingua::Ispell::pid,"\n";
  # system "ps axf |grep $Lingua::Ispell::pid";

  my %within;
  my $linenum = 0;
  my $total_words = 0;
  my $total_size = 0;
  my $prev_word = '';
  foreach my $line (split /\n/, $str) {
    $linenum++;
    # print "$linenum  $total_words  $total_size\r";
    $line =~ s{\\(begin|end)\{(.*?)\}}
              {$within{$2} = $1 eq 'begin';
               ' ' x length($&)
             }eg;
    next if $within{'tikzpicture'};
    next if $within{'align'};

    $total_size += length($line);
    foreach my $word (split /\s+/, $line) {
      next if $word eq '';  # split() unhelpfully gives empties
      $total_words++;
      # not macros, single letters, ending "," etc
      if ($word !~ /^\\|^.$|[[:punct:]]$/
          && $word eq $prev_word) {
        print "$filename:$linenum: doubled-word: $word\n";
        $exit_code = 1;
      }
      $prev_word = $word;
    }

    foreach my $r (Lingua::Ispell::spellcheck($line)) {
      # print "$r->{'type'}: $r->{'term'}\n";
      $exit_code = 1;
      if ($r->{'type'} eq 'miss') {
        my $misses = $r->{'misses'};
        if ($#$misses > 4) {
          $#$misses = 4;
          push @$misses, '...';
        }
        print "$filename:$linenum: $r->{term}",
          $misses ? join(' ','  maybe:',@$misses) : (), "\n";
      } elsif ($r->{'type'} eq 'guess') {
        my $guesses = $r->{'guesses'};
        print "$filename:$linenum: $r->{term}",
          $guesses ? join(' ','  maybe:',@$guesses) : (), "\n";
      } elsif ($r->{'type'} eq 'none') {
        print "$filename:$linenum: $r->{term}\n";
      }
    }
  }
  if ($option_verbose >= 1) {
    print "$filename total words $total_words\n";
  }
}

{
  my $help = sub {
    print "$FindBin::Script [--options] filename...\n";
    my @opts =
      (['--help, -?', 'Print this help'],
       ['--version',  'Print program version'],
       ['--verbose',  'Print extra messages'],
      );
    my $width = 2 + max (map { length ($_->[0]) } @opts);
    foreach (@opts) {
      printf "%-*s%s\n", $width, $_->[0], $_->[1];
    }
    print "\n";
    exit 0;
  };

  GetOptions ('help|?' => $help,
              version => sub {
                print "$FindBin::Script version $VERSION\n";
                exit 0;
              },
              'verbose:1' => \$option_verbose,
             )
    or exit 1;

  @ARGV or $help->();
}

foreach my $filename (@ARGV) {
  check_file ($filename);
}
exit $exit_code;


#------------------------------------------------------------------------------
__END__

=for stopwords Ryde ispell tex LaTeX

=head1 NAME

ispell-batch-tex.pl -- run ispell in batch mode on given TeX or LaTeX files

=head1 SYNOPSIS

 ispell-batch-tex.pl [--options] filename...

=head1 DESCRIPTION

Check spelling in the given TeX or LaTeX files using L<ispell(1)> and report
errors in batch style output like

    foo.tex:123: eroneous  maybe: erroneous

The first few of ispell's suggested corrections are shown (but not all since
on small words it tends to give a long list of rapidly decreasing
relevance).

Various TeX-isms are skipped for the checking, such as Latex package names
or everything in maths mode.  Ispell already has a TeX mode (and it it used
here), but further crunching here excludes more, and things across multiple
lines.  Most of the exclusions are LaTeX forms.

    % foo                     skip comments
    \foo                      all macro names skipped
    \label{foo}               labels, cites, etc, names
    \newtagform{foo}          amsmath tag form names
    \url{http://foo}          hyperref
    \href{http://foo}{...}    hyperref

Emacs style C<LocalWords:> lines are recognised and passed to ispell as
correct spellings, and which it can also offer back as suggestions.  Usual
TeX C<%> comment prefix can precede the C<LocalWords:> keyword.

    % LocalWords: jargonish unimprobableness

Doubled words like "theZ<> the" are reported similar to Emacs C<ispell.el>.
Any punctuation in between means not a double.  Single-letter doubles are
excluded, mainly since the parse can't recognise common acceptable cases
such as control letters of a C<tabular> environment.

The C<ispell> program is run via the L<Lingua::Ispell> module.  It expects
to find the program as C</usr/bin/ispell>.

=head1 OPTIONS

The command line options are

=over 4

=item C<--verbose>

=item C<--verbose=N>

Print more about what is done.  Verbosity level 0 (the default) is errors
only, 1 is some messages, and 2 is even more messages.

=item C<--help>

Print a brief help message.

=item C<--version>

Print the program version number and exit.

=back

=head1 OTHER WAYS TO DO IT

Emacs C<ispell.el> has some similar TeX crunching and exclusions (and with
regexp patterns to extend them if desired).  But C<M-x ispell-buffer> in a
big document can be quite slow.  C<ispell-batch-tex.pl> here can run in the
background and you can consider which reported spellings to fix or ignore
etc.

=head1 BUGS

Each file is slurped into memory in its entirety since this is easiest for
crunching multi-line forms to skip etc.  A large TeX file may use a lot of
memory to check, but if it's a size able to be edited then it should be
easily spell checked.

=head1 SEE ALSO

L<ispell(1)>,
L<Lingua::Ispell>

=head1 LICENSE

Copyright 2019, 2020 Kevin Ryde

ispell-batch-tex.pl is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 3, or (at your option) any later
version.

ispell-batch-tex.pl is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
ispell-batch-tex.pl.  If not, see L<http://www.gnu.org/licenses/>.

=cut
