#!/usr/bin/perl -w

# Copyright 2016, 2017 Kevin Ryde

# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


# Usage: perl anum-document.pl tex_filename pm_filename
#
# Compare A-numbers occurring in a .tex write-up versus a .pm PlanePath file.
# Most sequences mentioned in one should appear in the other, though the
# .tex might go into other territory, and the .pm might have mappings too
# obscure to appear in the .tex.

use 5.010;
use strict;
use warnings;
use File::Slurp;
use Math::OEIS::Names;
use Text::Tabs;
use Tie::IxHash;

my $names = Math::OEIS::Names->new;

sub pos_to_line_and_column {
  my ($str, $pos) = @_;
  $str = substr ($str, 0, $pos);
  my $nlpos = rindex ($str, "\n");
  my $lastline = substr ($str, $nlpos+1);
  $lastline = Text::Tabs::expand ($lastline);
  my $colnum = 1 + length ($lastline);
  my $linenum = 1 + scalar($str =~ tr/\n//);
  return ($linenum, $colnum);
}

# Return a hashref of $anum => [$linenum,$column] which is the A-numbers
# occurring in $filename, and the locations there.
# $samples is empty to match all A-numbers, or "_samples" to match only
# strings like A12345_samples.
#
sub filename_to_href {
  my ($filename, $samples) = @_;
  my $str = File::Slurp::read_file($filename)
    // die "Cannot open $filename: $!";
  my %hash;
  tie %hash, 'Tie::IxHash';
  my $re = qr/(A[0-9]+)$samples/;
  while ($str =~ /$re/g) {
    $hash{$1} //= [ pos_to_line_and_column($str,pos($str)) ];
  }
  my $count = scalar keys %hash;
  print "$filename: $count anums\n";
  return \%hash;
}

# Show the differences between A-numbers occurring in $filename1 and $filename2.
sub show_diffs {
  my ($filename1,$href1, $filename2,$href2) = @_;

  foreach my $anum (keys %$href1) {
    next if exists $href2->{$anum};
    my $anum_name = $names->anum_to_name($anum) // '[no name]';
    $anum_name =~ s/:/;/g;
    my ($linenum,$column) = @{$href1->{$anum}};
    print "$filename1:$linenum:$column: $anum $anum_name\n";
  }
}

sub anum_documents {
  my ($tex_filename,$pm_filename) = @_;
  my $tex_anums = filename_to_href($tex_filename,'_samples');
  my $pm_anums  = filename_to_href($pm_filename, '');

  show_diffs($tex_filename,$tex_anums,
             $pm_filename, $pm_anums);
  show_diffs($pm_filename, $pm_anums,
             $tex_filename,$tex_anums);
}
if (@ARGV) {
  while (@ARGV) {
    @ARGV >=2 or die "Odd number of command line filenames";
    anum_documents(splice @ARGV,0,2);
  }
} else {
  foreach my $elem
    (['../alt/alternate.tex',
      '../path/lib/Math/PlanePath/AlternatePaper.pm'
     ],
     ['../r5/r5dragon.tex',
      '../path/lib/Math/PlanePath/R5DragonCurve.pm'
     ],
     ['../ter/terdragon.tex',
      '../path/lib/Math/PlanePath/TerdragonCurve.pm'
     ],
     ['dragon.tex',
      '../path/lib/Math/PlanePath/DragonCurve.pm'
     ],
    ) {
    anum_documents(@$elem);
  }
}


#------------------------------------------------------------------------------
