# Copyright 2014, 2015, 2016, 2017, 2018, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


package MyFSM;
use 5.010;
use strict;
use warnings;
use Carp 'croak';
use List::Util 'max','all';
use Sort::Key::Natural 'natsort';
use Tie::IxHash;

# uncomment this to run the ### lines
# use Smart::Comments;

my @ipc;
END {
  foreach my $h (@ipc) {
    $h->finish;
  }
}

sub new {
  my $class = shift;
  my $self = bless { @_ }, $class;

  if (my $aref = delete $self->{'table_aref'}) {
    $self->{'table'} = { map {
      my $state = $_;
      my $row_aref = $aref->[$state];
      ### $row_aref
      $state => { map {$_ => $row_aref->[$_]} 0 .. $#$row_aref }
    } 0 .. $#$aref };
    ### table from aref: $self->{'table'}
  }

  $self->{'initials'} ||= {};
  if (defined(my $initial = $self->{'initial'})) {
    $self->{'initials'}->{$initial} = 1;
  }
  return $self;
}

sub states {
  my ($self) = @_;
  return keys %{$self->{'table'}};
}
sub has_state {
  my ($self, $state) = @_;
  return exists $self->{'table'}->{$state};
}
sub num_states {
  my ($self) = @_;
  return scalar(keys %{$self->{'table'}});
}
sub edges {
  my ($self) = @_;
  my ($state,$mapping) = each %{$self->{'table'}};
  return keys %$mapping;
}

sub sorted_states {
  my ($self) = @_;
  return $self->sort_states($self->states);
}
sub sorted_edges {
  my ($self) = @_;
  return natsort $self->edges;
}
sub sort_states {
  my $self = shift;
  my @states = natsort @_;
  @states = ((grep {  exists $self->{'initials'}->{$_}} @states),
             (grep {! exists $self->{'initials'}->{$_}} @states));
  return @states;
}

sub traverse {
  my ($self, $edges, $state) = @_;
  $state //= $self->{'initial'}
    // croak "No single initial state";
  my $table = $self->{'table'};
  foreach my $edge (@$edges) {
    $state = $table->{$state}->{$edge};
  }
  return $state;
}

sub show {
  my ($self) = @_;

  {
    my $table = $self->{'table'};
    my @states = $self->sorted_states;
    my @edges  = $self->sorted_edges;
    my $state_width = max(0, map {length} @states);
    my $edge_width  = max($state_width, map {length} @edges);

    printf "%-*s  ", $state_width, '';
    foreach my $edge (@edges) {
      printf " %-*s", $edge_width, $edge;
    }
    print "\n";

    foreach my $state (@states) {
      printf "%-*s  ", $state_width, $state;

      foreach my $edge (@edges) {
        my $destination = $table->{$state}->{$edge} // 'undef';
        printf " %-*s", $edge_width, $destination;
      }
      print "\n";
    }
  }
  {
    my @s;
    if (my @initials = $self->initial_states) {
      push @s, "initial(s)=" . join(',',@initials);
    }
    if (defined (my $flow = $self->{'flow'})) {
      push @s, "flow=$flow";
    }
    if (@s) {
      print join(',  ',@s),"\n";
    }
  }
  {
    my $accepting = $self->{'accepting'};
    if ($accepting && %$accepting) {
      my $prefix = 'accepting  ';
      my @states = $self->sort_states(keys %$accepting);
      my $state_width = max(0, map {length} @states);
      foreach my $state (@states) {
        printf "%s%*s = %s\n", $prefix, $state_width, $state, $accepting->{$state};
        $prefix = '        ';
      }
    }
  }

  print "total ",$self->num_states," states\n";
}

sub view {
  my ($self) = @_;
  my $graph = $self->as_graph_easy;
  my $str = $graph->as_graphviz;

  require File::Temp;
  my $t = File::Temp->new (UNLINK => 0, SUFFIX => '.ps');
  my $t_filename = $t->filename;
  require IPC::Run;
  IPC::Run::run(['dot','-Tps'], '<',\$str, '>',$t_filename);
  push @ipc, IPC::Run::start(['gv','--scale=1',$t_filename],'&');

  # Can gv read stdin?
  # require IPC::Run;
  # push @ipc, IPC::Run::start(['dot','-Tps'],'<',\$str,
  #                            '|',
  #                            ['gv','--scale=1','/dev/stdin']);
}
sub as_graphviz {
  my ($self) = @_;
  return $self->as_graph_easy->as_graphviz;
}
sub as_graph_easy {
  my ($self) = @_;
  require Graph::Easy;
  my $graph = Graph::Easy->new;
  $graph->{_warn_on_unknown_attributes} = 1;
  $graph->set_attribute('x-dot-overlap','false'); # neato
  $graph->set_attribute('x-dot-splines','true');  # neato

  my @edges = $self->sorted_edges;
  foreach my $state ($self->sorted_states) {
    print "as_graph_easy state $state\n";
    my %new_state_edges;
    foreach my $edge (@edges) {
      my $new_state = $self->transition($state,$edge);
      push @{$new_state_edges{$new_state}}, $edge;
    }
    while (my ($new_state, $edge_aref) = each %new_state_edges) {
      $graph->add_edge($state, $new_state, join(',', @$edge_aref));
    }
  }

  if (my $title = $self->{'title'}) {
    $graph->set_attribute('label',$title);
  }
  # print "flow is ",$graph->flow(),"\n";

  if (my $flow = $self->{'flow'}) {
    $graph->set_attribute('flow',$flow);
  }
  # print "flow is ",$graph->flow(),"\n";

  if (my @initials = $self->initial_states) {
    $graph->set_attribute('root',$initials[0]); # for as_graphviz()
    $graph->{att}->{root} = $initials[0];       # for root_node() for as_ascii()
    # ### root_node: $graph->root_node
  }

  return $graph;
}

sub transition {
  my ($self, $state,$edge) = @_;
  return $self->{'table'}->{$state}->{$edge}
    // croak "No such transition $state $edge";
}
sub add_transition {
  my ($self, $from,$edge,$to) = @_;
  if (exists $self->{'table'}->{$from}->{$edge}) {
    croak "Existing transition for $from $edge";
  }
  $self->{'table'}->{$from}->{$edge} = $to;
}

my $_periodically_prev_t = 0;
sub _periodically {
  my $t = time() & ~1;
  if ($t != $_periodically_prev_t) {
    $_periodically_prev_t = $t;
    return 1;
  } else {
    return 0;
  }
}

sub simplify {
  my ($self, %options) = @_;
  my $verbose = $options{'verbose'};
  my $max_equiv_count = $options{'max_equiv_count'};

  my $fsm = $self->{'table'};
  my $equiv_count = 0;
 EQUIV: for (;;) {
    last if defined $max_equiv_count && $equiv_count >= $max_equiv_count;
    my @states = $self->sorted_states;
    my @edges  = $self->sorted_edges;

    foreach my $state1 (@states) {
      foreach my $state2 (@states) {
        if (_periodically()) {
          local $| = 1;
          # print "consider $state1 and $state2\r";
        }
        my $equiv = $self->sub_equiv($state1,$state2) // next;

        if ($verbose) {
          my $num_equivs = scalar(keys %$equiv);
          print "EQ  $state1 = $state2  with $num_equivs others\n";
          my %printed;
          foreach my $state (sort keys %$equiv) {
            next if $printed{$state};
            my $other = $equiv->{$state};
            print "   $state  =  $other\n";
            $printed{$state} = 1;
            $printed{$other} = 1;
          }
        }

        if ($verbose) { print "   merge\n"; }
        _merge_equiv($self, $fsm, $equiv);
        if ($verbose) {
          print "------\n";
          print "merge to\n";
          $self->show;
        }
        $equiv_count++;
        next EQUIV;
      }
    }
    last;
  }

  $self->delete_unreached_states(%options);
}

sub accepting_states {
  my ($self) = @_;
  return keys %{$self->{'accepting'}};
}
sub non_accepting_states {
  my ($self) = @_;
  return grep {!$self->is_accepting($_)} $self->states;
}
sub initial_states {
  my ($self) = @_;
  ### $self
  return $self->sort_states(keys %{$self->{'initials'}});
}
sub set_initial_states {
  my $self = shift;
  return $self->{'initials'} = { map {$_=>1} @_ };
}

sub renumber_states {
  my ($self) = @_;
  my $n = 2;

  my @pending;
  if (my @initials = $self->initial_states) {
    @pending = @initials;
  } else {
    my @states = $self->sorted_states
      or return;  # no states at all
    @pending = ($states[0]);
  }

  my @edges  = $self->sorted_edges;
  my %target;
  my $temp_counter = 0;
  my %done;
  while (@pending) {
    my $state = shift @pending;
    my $new_state;
    if ($state eq $self->{'initial'}) {
      $new_state = 'initial';
      # $new_state = '1';
    } elsif ($state =~ /^(non,?)*$/) {
      $new_state = 'non';
    } elsif ($state =~ /^(enclosed,?)*$/) {
      $new_state = 'enclosed';
    } else {
      $new_state = $n++;
    }
    my $temp_state = "TEMPORARY--$temp_counter";
    $temp_counter++;
    $self->rename_state($state, $temp_state);
    $target{$temp_state} = $new_state;

    $done{$temp_state} = 1;
    foreach my $edge (@edges) {
      my $to_state = $self->{'table'}->{$temp_state}->{$edge};
      unless ($done{$to_state}++) {
        push @pending, $to_state;
      }
    }
  }
  while (my ($temp_state, $new_state) = each %target) {
    $self->rename_state($temp_state, $new_state);
  }
}

sub delete_unreached_states {
  my ($self, %options) = @_;
  my $verbose = $options{'verbose'};

  if ($verbose) {
    print "delete_unreached_states\n";
  }
  my %reached = %{$self->{'initials'}};
  my $more;
  do {
    $more = 0;
    foreach my $from_state ($self->states) {
      next unless exists $reached{$from_state};
      my $row = $self->{'table'}->{$from_state};
      foreach my $to_state (values %$row) {
        next if exists $reached{$to_state};
        if ($verbose) { print "  $to_state reached from $from_state\n"; }
        $reached{$to_state} = 1;
        $more = 1;
      }
    }
  } while ($more);

  my $deleted = 0;
  foreach my $state (keys %{$self->{'table'}}) {
    next if $reached{$state};
    if ($verbose) { print "  unreached $state\n"; }
    delete $self->{'table'}->{$state};
    delete $self->{'accepting'}->{$state};
    $deleted++;
  }
  if ($verbose) { print "  deleted $deleted states\n"; }
}

# return true if $state1 and $state2 might be treated as equivalent
sub _equiv_possible {
  my ($self, $state1,$state2) = @_;
  if ($state1 eq $state2) {
    ## equiv of same state ...
    return 1;
  }

  # same accepting type for both
  my $accepting1 = $self->{'accepting'}->{$state1};
  my $accepting2 = $self->{'accepting'}->{$state2};
  return ((defined $accepting1 && defined $accepting2
           && $accepting1 eq $accepting2)
          || (! defined $accepting1 && ! defined $accepting2));
}

# return a hashref of equivalences ...
sub sub_equiv {
  my ($self, $state1,$state2) = @_;
  ### sub_equiv: "$state1  $state2"

  if ($state1 eq $state2
      || ! _equiv_possible($self, $state1,$state2)) {
    return undef;
  }

  my $table = $self->{'table'};
  my %equiv;
  tie %equiv, 'Tie::IxHash';
  %equiv = ($state1 => $state1,
            $state2 => $state1);

  my $prev_num_equiv = 0;
  while (scalar(keys %equiv) != $prev_num_equiv) {
    $prev_num_equiv = scalar(keys %equiv);

    my @pending = keys %equiv;
    while (@pending) {
      my $state1 = shift @pending;
      my $state2 = $equiv{$state1};

      my $mapping1 = $table->{$state1};
      my $mapping2 = $table->{$state2};

      ### check outputs of: "$state1  $state2"

      foreach my $edge ($self->sort_states(keys %$mapping1)) {
        my $new_state1 = $mapping1->{$edge};
        my $new_state2 = $mapping2->{$edge};
        ### edge: "$edge  $new_state1 $new_state2"

        if ($new_state1 eq $new_state2) {
          ### same output, good ...
          next;
        }
        if (! _equiv_possible($self, $new_state1,$new_state2)) {
          ### outputs cannot be made equivalent ...
          return undef;
        }

        if (exists $equiv{$new_state1} && exists $equiv{$new_state2}) {
          if ($equiv{$new_state1} eq $equiv{$new_state2}) {
            # same outputs under equivalences, good
            next;
          } else {
            # have proposed equivalence, and the outputs are not the same
            return undef;
          }
        }

        if (exists $equiv{$new_state1}) {
          # propose as new equivalence
          $equiv{$new_state2} = $equiv{$new_state1};

        } elsif (exists $equiv{$new_state2}) {
          # propose as new equivalence
          $equiv{$new_state1} = $equiv{$new_state2};

        } else {
          # propose as new equivalence
          $equiv{$new_state1} = $new_state1;
          $equiv{$new_state2} = $new_state1;
        }
      }
    }
  }
  return \%equiv;
}

sub _merge_equiv {
  my ($self, $fsm, $equiv) = @_;
  my $initials = $self->{'initials'};
  foreach my $group (_equiv_hash_to_groups($equiv)) {
    my @states = $self->sort_states(@$group);
    my $state1 = shift @states;
    foreach my $state2 (@states) {
      # print "   merge $state1 = $state2\n";
      if (exists $initials->{$state2} && ! exists $initials->{$state1}) {
        ($state1,$state2) = ($state2,$state1);
      }
      $state1 = $self->merge_state($state1,$state2);
    }
  }
}
sub _equiv_hash_to_groups {
  my ($equiv) = @_;
  my %groupnum;
  my $upto = 1;
  foreach my $state1 (keys %$equiv) {
    my $state2 = $equiv->{$state1};
    if (my $g1 = $groupnum{$state1}) {
      if (my $g2 = $groupnum{$state2}) {
        # group merger
        foreach my $s (keys %groupnum) {
          if ($groupnum{$s} == $g2) {
            $groupnum{$s} = $g1;
          }
        }
      } else {
        $groupnum{$state2} = $g1;
      }
    } else {
      if (my $g2 = $groupnum{$state2}) {
        $groupnum{$state1} = $g2;
      } else {
        $groupnum{$state1} = $groupnum{$state2} = $upto++;
      }
    }
  }
  my %groups;
  foreach my $s (keys %groupnum) {
    push @{$groups{$groupnum{$s}}}, $s;
  }
  return values %groups;
}

# Merge $state1 and $state2.
# Any transitions going to $state2 are changed to go to $state1.
sub merge_state {
  my ($self, $state1, $state2, %options) = @_;
  ### merge_state: "$state1, $state2"
  if ($state1 eq $state2) { return; }
  my $table = $self->{'table'};
  delete $table->{$state2};
  delete $self->{'accepting'}->{$state2};
  foreach my $mapping (values %$table) {
    foreach my $new_state (values %$mapping) {
      if ($new_state eq $state2) {
        $new_state = $state1;  # mutate $table
      }
    }
  }
  my $new_name = $options{'name'} // "$state1 and $state2";
  $self->rename_state($state1, $new_name);
  return $new_name;
}

sub state_is_accepting {
  my ($self, $state) = @_;
  return defined $self->{'accepting'}->{$state};
}
sub add_accepting {
  my ($self, $state, $type) = @_;
  if (! defined $type) { $type = 1; }
  $self->{'accepting'}->{$state} = $type;
}

my %flow_reverse = (east => 'west', west => 'east',
                    north => 'south', south => 'north',
                   );
sub reverse {
  my ($self) = @_;
  my @states = $self->states;
  my @edges  = $self->edges;
  my %edge_mappings;
  foreach my $state (@states) {
    foreach my $edge (@edges) {
      $edge_mappings{$edge}->{$state} = $self->transition($state,$edge);
    }
  }
  my $identity = { map {$_ => $_} @states };
  my $initial = $self->mapping_name($identity);
  ### $initial

  my @pending = ($identity);
  my %mapping_seen;
  my %table;
  my %accepting;
  while (@pending) {
    my $mapping = pop @pending;
    my $mapping_name = $self->mapping_name($mapping);

    if (defined(my $accepting_name
                = $self->mapping_is_all_accepting($mapping))) {
      $accepting{$mapping_name} = $accepting_name;
    }

    foreach my $edge (@edges) {
      my $product = _mapping_product($edge_mappings{$edge}, $mapping);
      my $product_name = $self->mapping_name($product);
      $table{$mapping_name}->{$edge} = $product_name;
      if (! $mapping_seen{$product_name}++) {
        push @pending, $product;
      }
    }
  }
  ### %table

  my $flow = $self->{'flow'} || 'east';
  $flow = $flow_reverse{$flow} || $flow;

  my $title = join(', ',
                   ($self->{'title'} // ()),
                   'reverse');

  my $class = ref $self;
  my $rev = $class->new (table     => \%table,
                         initial   => $initial,
                         accepting => \%accepting,
                         title     => $title,
                         flow      => $flow,
                        );
  $rev->rename_state($initial, 'identity');
  return $rev;
}
sub mapping_name {
  my ($self, $mapping) = @_;
  if (defined (my $a = $self->mapping_is_all_accepting($mapping))) {
    return $a;
  }
  # if ($self->mapping_is_all_non_accepting($mapping)) {
  #   return 'non';
  # }
  if ($self->mapping_is_identity($mapping)) {
    return 'identity';
  }
  return join(',', map {$mapping->{$_}} $self->sorted_states);
}
sub mapping_is_identity {
  my ($self, $mapping) = @_;
  while (my ($state, $new_state) = each %$mapping) {
    if ($state ne $new_state) { return 0; }
  }
  return 1;
}
sub mapping_is_all_accepting {
  my ($self, $mapping) = @_;
  my $accepting_name;
  foreach my $new_state (values %$mapping) {
    my $a = $self->{'accepting'}->{$new_state};
    $accepting_name //= $a;
    if (! defined $a || $a ne $accepting_name) {
      return undef;
    }
  }
  return $accepting_name;
}
sub mapping_is_all_non_accepting {
  my ($self, $mapping) = @_;
  return all {!$self->is_accepting($_)} values %$mapping;
}
sub _mapping_product {
  my ($m1, $m2) = @_;
  my %ret;
  foreach my $state (keys %$m1) {
    $ret{$state} = $m2->{$m1->{$state}};
  }
  return \%ret;
}

sub is_accepting {
  my ($self, $state) = @_;
  return defined($self->{'accepting'}->{$state});
}

sub rename_state {
  my ($self, $old, $new) = @_;
  if ($old eq $new) {
    return;   # no change
  }
  my $table = $self->{'table'};
  if (! $table->{$old}) {
    croak "No such state $old";
  }
  if ($table->{$new}) {
    croak "Rename would overwrite state new";
  }
  $table->{$new} = delete $table->{$old};

  if (exists $self->{'accepting'}->{$old}) {
    $self->{'accepting'}->{$new} = delete $self->{'accepting'}->{$old};
  }
  if (exists $self->{'initials'}->{$old}) {
    $self->{'initials'}->{$new} = delete $self->{'initials'}->{$old};
  }
  if (defined $self->{'initial'} && $self->{'initial'} eq $old) {
    $self->{'initial'} = $new;
  }
  foreach my $mapping (values %$table) {
    foreach my $destination (values %$mapping) {
      if ($destination eq $old) {
        $destination = $new;  # mutate $table
      }
    }
  }
}

sub union {
  my ($f1, $f2, %options) = @_;

  my $title;
  {
    my @title = (($f1->{'title'} // ()),
                 ($f2->{'title'} // ()));
    if (@title) {
      $title = join(' and ', @title);
    }
  }

  my %table;
  my @f1_states = $f1->states;
  my @f2_states = $f2->states;
  my @edges = $f1->edges;
  foreach my $edge (@edges) {
    foreach my $s1 (@f1_states) {
      my $new_s1 = $f1->transition($s1,$edge);
      foreach my $s2 (@f2_states) {
        my $new_s2 = $f2->transition($s2,$edge);
        $table{"$s1+$s2"}->{$edge} = "$new_s1+$new_s2";
      }
    }
  }

  my %accepting;
  if (($options{'type'}||'') eq 'either') {
    foreach my $s1 ($f1->accepting_states) {
      foreach my $s2 ($f2->states) {
        $accepting{"$s1+$s2"} = 'either'; # f1 '.$f1->{'accepting'}->{$s1};
      }
    }
    foreach my $s1 ($f1->states) {
      foreach my $s2 ($f2->accepting_states) {
        $accepting{"$s1+$s2"} = 'either'; # 'f2 '.$f2->{'accepting'}->{$s2};
      }
    }
    $title .= ', either';

  } elsif (($options{'type'}||'') eq 'merge') {
    foreach my $s1 ($f1->accepting_states) {
      foreach my $s2 ($f2->states) {
        $accepting{"$s1+$s2"} = 'f1 '.$f1->{'accepting'}->{$s1};
      }
    }
    foreach my $s1 ($f1->states) {
      foreach my $s2 ($f2->accepting_states) {
        $accepting{"$s1+$s2"} = 'f2 '.$f2->{'accepting'}->{$s2};
      }
    }
    $title .= ', merge';

  } else {
    foreach my $s1 ($f1->accepting_states) {
      foreach my $s2 ($f2->accepting_states) {
        $accepting{"$s1+$s2"} =
          $f1->{'accepting'}->{$s1} . '+' . $f2->{'accepting'}->{$s2};
      }
    }
    $title .= ', both';
  }

  my %initials;
  foreach my $s1 ($f1->initial_states) {
    foreach my $s2 ($f2->initial_states) {
      $initials{"$s1+$s2"} =
        $f1->{'initials'}->{$s1} . '+' . $f2->{'initials'}->{$s2};
    }
  }

  my $initial;
  if (defined (my $s1 = $f1->{'initial'})) {
    if (defined (my $s2 = $f2->{'initial'})) {
      $initial = "$s1+$s2";
    }
  }

  my $class = ref $f1;
  return $class->new (table     => \%table,
                      initial   => $initial,
                      accepting => \%accepting,
                      initials  => \%initials,
                      title     => $title,
                      flow      => $f1->{'flow'});
}

sub check {
  my ($self) = @_;
  my $sep = "\n";
  my $bad = 0;
  foreach my $state ($self->initial_states) {
    if (! defined $self->{'table'}->{$state}) {
      print "${sep}initial $state  not in table\n";
      $sep = '';
      $bad++;
    }
  }
  foreach my $state ($self->accepting_states) {
    if (! defined $self->{'table'}->{$state}) {
      print "${sep}accepting $state  not in table\n";
      $sep = '';
      $bad++;
    }
  }
  my @edges = $self->edges;
  my %edges = map {$_ => 1} @edges;
  foreach my $state ($self->states) {
    my $row = $self->{'table'}->{$state};
    foreach my $edge (@edges) {
      if (! defined $row->{$edge}) {
        print "${sep}state $state no edge $edge\n";
        $sep = '';
        $bad++;
      }
    }
    foreach my $edge (keys %$row) {
      if (! defined $edges{$edge}) {
        print "${sep}state $state extra edge $edge\n";
        $sep = '';
        $bad++;
      }
    }
  }
  $| = 1;
  if ($bad) {
    croak "Bad";
  }
}

sub digit_delta {
  my ($self) = @_;
  ### digit_delta() ...

  my @states = $self->states;
  my @edges = $self->sorted_edges;

  my %table;
  foreach my $from_state (@states) {
    foreach my $prev_edge_index (0 .. $#edges) {
      my $prev_edge = $edges[$prev_edge_index];
      my $from = "$from_state-prev$prev_edge";
      foreach my $offset (0 .. $#edges) {
        my $edge_index = ($prev_edge_index + $offset) % scalar(@edges);
        my $edge = $edges[$edge_index];
        my $to_state = $self->{'table'}->{$from_state}->{$edge};
        my $to = "$to_state-prev$edge";
        $table{$from}->{"+$offset"} = $to;
      }
    }
  }

  my $initial;
  if (defined $self->{'initial'}) {
    $initial = "$self->{'initial'}-prev$edges[0]";
  }

  my %accepting;
  while (my ($state, $accept) = each %{$self->{'accepting'}}) {
    foreach my $edge (@edges) {
      $accepting{"$state-prev$edge"} = $accept;
    }
  }

  my $title = join(', ',
                   ($self->{'title'} // ()),
                   'delta');

  my $class = ref $self;
  my $delta = $class->new (table     => \%table,
                           initial   => $initial,
                           accepting => \%accepting,
                           title     => $title,
                           flow      => $self->{'flow'},
                          );
  ### $delta
  $delta->delete_unreached_states;
  return $delta;
}

#------------------------------------------------------------------------------

sub _to_tikz_node_name {
  my ($str) = @_;
  $str =~ tr/,/-/;
  return $str;
}
sub print_tikz {
  my ($self) = @_;

  my @states = $self->sorted_states;
  my @edges = $self->sorted_edges;
  my $flow = $self->{'flow'} || 'east';

  my $rows = int(sqrt(scalar(@states)));
  my $r = 0;
  my $c = 0;
  foreach my $state (@states) {
    my $x = ($flow eq 'west' ? -$c : $c);
    my $name = _to_tikz_node_name($state);
    print "  \\node ($name) at ($x,$r) [my box] {$state};\n";
    $r++;
    if ($r >= $rows) {
      $c++;
      $r = 0;
    }
  }
  print "\n";

  foreach my $from_state (@states) {
    print "  % $from_state\n";
    my %to;
    tie %to, 'Tie::IxHash';
    foreach my $edge (@edges) {
      my $to_state = $self->{'table'}->{$from_state}->{$edge};
      push @{$to{$to_state}}, $edge;
    }
    foreach my $to_state (keys %to) {
      my $edges = join(',', @{$to{$to_state}});
      my $from_name = _to_tikz_node_name($from_state);
      if ($from_state eq $to_state) {
        print "  \\draw [->,loop below] ($from_name) to node[pos=.12,auto=left] {$edges} ();\n";
      } else {
        my $to_name = _to_tikz_node_name($to_state);
        my $bend = $self->has_edge($to_state,$from_state)
          ? ',bend left=10' : '';
        print "  \\draw [->$bend] ($from_name) to node[pos=.5,auto=left] {$edges} ($to_name);\n";
      }
    }
    print "\n";
  }
}

# Return true if there is any transition from $from_state to $to_state.
sub has_edge {
  my ($self, $from_state, $to_state) = @_;
  my $row = $self->{'table'}->{$from_state};
  while (my ($edge,$to) = each %$row) {
    if ($to eq $to_state) { return 1; }
  }
  return 0;
}

#------------------------------------------------------------------------------
sub as_sparse6 {
  my $self = shift;
  $self->as_graph6(@_, format => 'sparse6');
}

sub as_graph6 {
  my ($self, %options) = @_;

  my @edges = $self->edges;
  my @states = $self->sorted_states;
  my %state_to_num = map {$states[$_] => $_} 0 .. $#states;
  my @graph_edges;
  foreach my $from_state (@states) {
    my $from_num = $state_to_num{$from_state};
    foreach my $edge (@edges) {
      my $to_state = $self->{'table'}->{$from_state}->{$edge};
      my $to_num = $state_to_num{$to_state};
      push @graph_edges, [$from_num, $to_num];
    }
  }

  require Graph::Graph6;
  Graph::Graph6::write_graph(%options,
                             str_ref => \my $str,
                             edge_aref => \@graph_edges,);
  return $str;
}

#------------------------------------------------------------------------------

sub accepting_sequences {
  my ($self, $k) = @_;
  my @edges = $self->sorted_edges;
  my $table = $self->{'table'};

  my @ret;
  my @edge_index = (-1);
  my @edge_sequence;
  my @state_sequence = ($self->{'initial'});

  my $pos = 0;
  while ($pos >= 0) {
    ### at: "$pos edge_index $edge_index[$pos]"

    if ($pos >= $k) {
      ### limit ..
      # print join(' ',@state_sequence),"\n";
      # print "E ",join(' ',@edge_sequence),"\n";
      if ($self->is_accepting($state_sequence[$pos])) {
        push @ret, [ @edge_sequence ];
      }
      $pos--;
      next;
    }
    $edge_index[$pos]++;
    if ($edge_index[$pos] > $#edges) {
      $pos--;
      next;
    }
    my $edge = $edges[$edge_index[$pos]];
    $edge_sequence[$pos] = $edge;

    my $from_state = $state_sequence[$pos];
    my $to_state = $table->{$from_state}->{$edge};

    $state_sequence[$pos+1] = $to_state;
    $edge_index[$pos+1] = -1;
    $pos++;
  }
  return @ret;
}

#------------------------------------------------------------------------------

sub eventually_accepting {
  my ($self, %options) = @_;
  my $more = 1;
  while ($more) {
    $more = 0;
    my @edges = $self->sorted_edges;
    my @states = $self->sorted_states;
    my $table = $self->{'table'};
    my $accepting = $self->{'accepting'};

  STATE: foreach my $state (@states) {
      next if $accepting->{$state};

      my $row = $self->{'table'}->{$state};
      foreach my $edge (@edges) {
        my $to_state = $row->{$edge};
        if (! defined $accepting->{$to_state}) {
          next STATE;
        }
      }

      $more = 1;
      my $to_state = $row->{$edges[0]};
      $accepting->{$state} = $accepting->{$to_state};
      if ($options{'verbose'}) {
        print "eventually_accepting $state\n";
      }
      if ($options{'merge'}) {
        $self->merge_state($to_state, $state, name => $to_state);
        last;
      }
    }
  }
}

#------------------------------------------------------------------------------
# FLAT

sub as_FLAT {
  my ($self, %options) = @_;
  my @edges = $self->sorted_edges;
  my @states = $self->sorted_states;
  my $table = $self->{'table'};

  my $flat = FLAT::DFA->new;
  my @flat_states = $flat->add_states(scalar(@states));
  my %state_to_flat_state
    = (map {$states[$_] => $flat_states[$_]} 0 .. $#states);

  foreach my $state (@states) {
    my $row = $self->{'table'}->{$state};
    foreach my $edge (@edges) {
      my $to_state = $row->{$edge};
      $flat->add_transition($state_to_flat_state{$state},
                            $state_to_flat_state{$to_state},
                            $edge);
    }
  }
  $flat->set_accepting (map {$state_to_flat_state{$_}}
                        keys %{$self->{'accepting'}});
  $flat->set_starting(map {$state_to_flat_state{$_}} $self->initial_states);
  return $flat;
}

#------------------------------------------------------------------------------

1;
__END__
