#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use Graph::Maker::TwindragonAreaTree;
use Math::PlanePath::ComplexPlus;
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'mul','div','chamfer','xy_rotate_plus90';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 6;  # of twindragon

print "  % generated by tools/$FindBin::Script\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

my $path = Math::PlanePath::DragonCurve->new;
my $complex_plus = Math::PlanePath::ComplexPlus->new;

my ($n_lo, $n_hi) = $path->level_to_n_range($level+1);
my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
### level: "$level  n_hi=$n_hi"

sub xyxy_is_twindragon_touch {
  my ($x,$y) = @_;
  ### xyxy_is_twindragon_touch(): "$x, $y"

  my @n_list = ($path->xy_to_n_list($x,$y),
                $path->xy_to_n_list($x_hi-$x,$y_hi-$y));
  @n_list = grep {$_ < $n_hi} @n_list;
  return @n_list > 1;
}

$Printwrap = 2;
{
  my ($x,$y) = $path->n_to_xy($n_lo);
  printwrap("\\node (start) at ($x,$y) [my dot];\n");
}
{
  my ($x,$y) = $path->n_to_xy($n_hi);
  printwrap("\\node (end) at ($x,$y) [my dot];\n");
}
{
  my ($x,$y) = $path->n_to_xy($n_hi*3/4);
  printwrap("\\coordinate (middle) at ($x,$y);\n");
}

# twindragon drawing
{
  $Printwrap = 2;
  printwrap("\\draw[my dragon line] ");
  my $join = '';
  $Printwrap = 4;
  foreach my $rev (0,1) {
    foreach my $n ($n_lo .. $n_hi-1) {

      my ($x,$y) = $path->n_to_xy($n);
      if ($rev) {
        ($x,$y) = ($x_hi-$x, $y_hi-$y);
      }

      if (xyxy_is_twindragon_touch($x,$y)) {
        my ($sdx,$sdy) = $path->n_to_dxdy($n-1);
        my ($tdx,$tdy) = $path->n_to_dxdy($n);
        if ($rev) {
          ($sdx,$sdy) = (-$sdx,-$sdy); # rotate 180
          ($tdx,$tdy) = (-$tdx,-$tdy); # rotate 180
        }
        my $sx = $x . chamfer(-$sdx);
        my $sy = $y . chamfer(-$sdy);
        my $tx = $x . chamfer($tdx);
        my $ty = $y . chamfer($tdy);
        printwrap("$join($sx,$sy)");
        printwrap("--($tx,$ty)");
      } else {
        printwrap("$join($x,$y)");
      }
      $join = '--';
    }
  }
  printwrap("--cycle;\n");
}

# twindragon area tree
{
  my $tree_segments = MyLineSegments->new;
  my $graph = Graph::Maker->new ('twindragon_area_tree',
                                 level => $level,
                                 undirected => 1);
  foreach my $edge ($graph->edges) {
    $tree_segments->add(map {$complex_plus->n_to_xy($_)} @$edge);
  }

  my $transform = sub {
    my ($x,$y) = @_;
    ($x,$y) = div($x,$y);
    $x = -$x; $y = -$y;  # rotate 180
    $x += .5; $y += .5;
    return ($x,$y);
  };
  $Printwrap = 2;
  printwrap("\\draw[my tree line] ");
  $Printwrap += 2;
  $tree_segments->printwrap(transform => $transform);
  printwrap(";\n");

  $Printwrap = 2;
  printwrap("\\path foreach \\p in {");
  $Printwrap = 4;
  foreach my $n (sort $graph->vertices) {
    my ($x,$y) = $complex_plus->n_to_xy($n);
    ($x,$y) = $transform->($x,$y);
    printwrap("($x,$y),");
  }
  printwrap("}");
  printwrap(" { \\p node[my vertex dot] };\n");

  my $diameter = $graph->diameter;
  print "  \\newcommand{\\MyDiameter}{$diameter}\n";
  print "% GP-Test-Last  TAdiameter($level) == $diameter\n";
  print "% GP-Test  LQ($level+1)-1 == $diameter\n";
}

print "  % end generated\n";
exit 0;
