#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2018, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use List::Util 'min';
use Math::PlanePath::DragonCurve 118; # v.118 for xyxy_to_n_either()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'BlobN','chamfer','xy_rotate_plus90','xy_rotate_minus90';
use MyPrintwrap ':all';


my $level = 8;

print "  % generated by tools/$FindBin::Script\n";
my $path = Math::PlanePath::DragonCurve->new;
my ($n_lo, $n_hi) = $path->level_to_n_range($level);

sub path_xyxy_is_side_enclosed {
  my ($path, $x1,$y1, $x2,$y2, $side, $n_limit) = @_;
  if (! path_xyxy_is_traversed_either($path, $x1,$y1, $x2,$y2, $n_limit)) {
    return 0;
  }
  my $dx = $x2 - $x1;
  my $dy = $y2 - $y1;
  if ($side eq 'left') {
    ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # to left side
  } else {
    ($dx,$dy) = xy_rotate_minus90($dx,$dy);  # to right side
  }
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  if (path_xyxy_is_traversed_either($path, $x1,$y1, $x1s,$y1s, $n_limit)
      && path_xyxy_is_traversed_either($path, $x2,$y2, $x2s,$y2s, $n_limit)
      && path_xyxy_is_traversed_either($path, $x1s,$y1s, $x2s,$y2s, $n_limit)) {
    my @array = ($x1,$y1, $x2,$y2, $x2s,$y2s, $x1s,$y1s);
    for (;;) {
      my $mx = min($array[0], $array[2], $array[4], $array[6]);
      my $my = min($array[1], $array[3], $array[5], $array[7]);
      last if $mx == $array[0] && $my == $array[1];
      push @array, splice @array,0,2;
    }
    return \@array;
  } else {
    return '';
  }
}

sub path_xyxy_is_traversed_either {
  my ($path, $x1,$y1, $x2,$y2, $n_limit) = @_;
  my $n = $path->xyxy_to_n_either($x1,$y1, $x2,$y2);
  return defined $n && $n < $n_limit;
}

sub is_touch {
  my ($n) = @_;
  my ($x,$y) = $path->n_to_xy($n);
  foreach my $n_other ($path->xy_to_n_list($x,$y)) {
    next if $n_other == $n;
    next if $n_other > $n_hi;
    return 1;
  }
  return 0;
}

my $n_blob_start = BlobN($level);
my $n_blob_end = BlobN($level+1) - 3;

my $n_draw_start = $n_blob_start-1;
my $n_draw_end = $n_blob_end+1;

$Printwrap = 2;
{
  my ($x,$y) = $path->n_to_xy($n_draw_start);
  printwrap("\\coordinate (start) at ($x,$y);\n");
}
{
  my ($x,$y) = $path->n_to_xy($n_draw_end);
  printwrap("\\coordinate (end) at ($x,$y);\n");
}

printwrap("\\draw ");
$Printwrap = 4;
my $join = '';
foreach my $n ($n_draw_start .. $n_draw_end) {
  my ($x,$y) = $path->n_to_xy($n);
  my ($dx,$dy) = $path->n_to_dxdy($n);

  if ($n != $n_draw_start && $n != $n_draw_end && is_touch($n,$n_blob_end)) {
    my ($prev_dx,$prev_dy) = $path->n_to_dxdy($n-1);
    my $sx = $x . chamfer(-$prev_dx);
    my $sy = $y . chamfer(-$prev_dy);
    my $tx = $x . chamfer($dx);
    my $ty = $y . chamfer($dy);
    printwrap("$join($sx,$sy)");
    printwrap("--($tx,$ty)");
  } else {
    printwrap("$join($x,$y)");
  }
  $join = '--';
}
printwrap(";\n");

my %colour = (left => '[my grey]', right => '');
my %count = (left => 0, right => 0);
foreach my $side ('left','right') {
  $Printwrap = 2;
  printwrap("\\fill$colour{$side} foreach \\p in {");
  $Printwrap = 4;
  my %seen;
  foreach my $n ($n_blob_start .. $n_blob_end-1) {
    my ($x,$y) = $path->n_to_xy($n);
    my ($dx,$dy) = $path->n_to_dxdy($n);

    my $aref = path_xyxy_is_side_enclosed($path, $x,$y, $x+$dx,$y+$dy, $side, $n_hi)
      || next;
    my ($x1,$y1, $x2,$y2, $x3,$y3, $x4,$y4) = @$aref;
    next if $seen{"$x1,$y1"}++;
    printwrap("($x1,$y1),");
    $count{$side}++;
  }
  printwrap("}");
  printwrap(" { \\p +(\\MyMargin,\\MyMargin) rectangle +(1-\\MyMargin,1-\\MyMargin) };\n");
}

my $BlobAL = $count{'left'};
my $BlobAR = $count{'right'};
my $BlobA = $BlobAL + $BlobAR;
print "  \\newcommand{\\MyLevel}{$level}\n";
print "  \\newcommand{\\MyBlobAL}{$BlobAL}\n";
print "  \\newcommand{\\MyBlobAR}{$BlobAR}\n";
print "  \\newcommand{\\MyBlobA}{$BlobA}\n";
print "% GP-Test-Last  BlobAL($level)==$BlobAL\n";
print "% GP-Test-Last  BlobAR($level)==$BlobAR\n";
print "% GP-Test-Last  BlobA($level)==$BlobA\n";
print "  % end generated\n";
exit 0;
