#!/usr/bin/perl -w

# Copyright 2016, 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use List::Util 'min','max';
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'BlobN','BlobSegments','BlobEnd_xy',
  'points_to_hull','hull_points_sort_by_angle','transform_to_horizontal';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 15;
if (@ARGV) { $level = shift @ARGV; }

print "  % generated by tools/$FindBin::Script\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

my $path = Math::PlanePath::DragonCurve->new;
my ($n_lo,$n_hi) = $path->level_to_n_range($level);
my $blob_n_lo = BlobN($level);
my $blob_n_hi = BlobN($level+1)-3;

sub transform {
  my ($x,$y) = @_;
  return transform_to_horizontal($x,$y, $level);
}

if (0) {
  # dragon curve lines
  my $segments = MyLineSegments->new;
  foreach my $n ($blob_n_lo .. $blob_n_hi-1) {
    my ($x,$y) = $path->n_to_xy($n);
    $segments->add($x,$y, $path->n_to_xy($n + 1));
  }

  $Printwrap = 2;
  printwrap("\\draw[my dragon line] ");
  $Printwrap += 2;
  $segments->printwrap (transform => \&transform);
  printwrap(";\n");
}

my $b4 = BlobSegments($level-4);
my $b2 = BlobSegments($level-2);

my @type;
foreach my $n ($blob_n_lo .. $blob_n_hi) {    # blob k-2 segments
  $type[$n] = 'blob-2';
}

my $b4_first_lo = $blob_n_lo;
my $b4_first_hi = $b4_first_lo + $b4;

my $b4_last_hi = $blob_n_hi;
my $b4_last_lo = $b4_last_hi - $b4;

my $b2_lo = $b4_first_hi + 3;
my $b2_hi = $b4_last_lo - 3;

# BlobPartsTC(11) == -8 + 24*I
# dragon_unpoint(-8 + 24*I,  1) == [0, 1408]
# dragon_unpoint(-8 + 24*I, -1) == [0, 896]
#  896/2^11 ==  7/16
# 1408/2^11 == 11/16
my $twin_lo = $n_hi *  7/16;
my $twin_hi = $n_hi * 11/16;

foreach my $n ($b4_first_lo .. $b4_first_hi) {   # blob k-4 first segments
  $type[$n] = 'blob-4a';
}
foreach my $n ($b4_first_hi+1 .. $b2_lo-1) {   # bridge segments
  $type[$n] = 'bridge';
}
foreach my $n ($b4_last_lo .. $b4_last_hi) {
  $type[$n] = 'blob-4b';
}
foreach my $n ($b4_last_lo-2 .. $b4_last_lo-1) {   # bridge segments
  $type[$n] = 'bridge';
}
foreach my $n ($twin_lo .. $twin_hi-1) {    # twindragon
  $type[$n] = 'twindragon';
}

{
  my ($x,$y) = $path->n_to_xy($blob_n_lo);
  ($x,$y) = transform($x,$y);
  print "  \\coordinate (start) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($blob_n_hi);
  ($x,$y) = transform($x,$y);
  print "  \\coordinate (end) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($b4_first_hi);
  ($x,$y) = transform($x,$y);
  print "  \\coordinate (blob-4a-end) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($b2_lo);
  ($x,$y) = transform($x,$y);
  print "  \\coordinate (blob-2-end) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($b2_hi);
  ($x,$y) = transform($x,$y);
  print "  \\coordinate (blob-2-start) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($b4_last_lo);
  ($x,$y) = transform($x,$y);
  print "  \\coordinate (blob-4b-start) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($twin_lo + 3*2**($level-4));
  ($x,$y) = transform($x,$y);
  print "  \\coordinate (twindragon-start) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($twin_lo + 2**($level-4));
  ($x,$y) = transform($x,$y);
  print "  \\coordinate (twindragon-end) at ($x,$y);\n";
}

my %points;
foreach my $n ($blob_n_lo .. $blob_n_hi) {
  my ($x,$y) = $path->n_to_xy($n);
  push @{$points{$type[$n]}}, [$x,$y];
}

my %type_to_P = ('blob-4a' => 'BP',
                 'blob-4b' => 'BP',
                 'blob-2'  => 'BP',
                 'twindragon' => 'P');
my %type_to_Psuffix = ('blob-4a' => 'a',
                       'blob-4b' => 'b',
                       'blob-2'  => 'm',
                       'twindragon' => 't');
my %type_to_level_offset = ('blob-4a' => 4,
                            'blob-4b' => 4,
                            'blob-2'  => 2,
                            'twindragon' => 4);
foreach my $type ('blob-4a', 'blob-4b', 'blob-2', 'twindragon') {
  my $points = $points{$type} || die;
  $points = points_to_hull($points);
  ### $points

  my $transformed_points = [ map {[transform(@$_)]} @$points ];
  ### $transformed_points
  my $P1_y = min(map {$_->[1]} @$transformed_points);
  my $P1_x = max(map {$_->[0]} grep {$_->[1]==$P1_y} @$transformed_points);
  my ($P1_i) = grep {$transformed_points->[$_]->[0] == $P1_x
                       && $transformed_points->[$_]->[1] == $P1_y
                     } 0 .. $#$transformed_points
                     or die;
  if ($type eq 'blob-2') { $P1_i += 5; }
  if ($type eq 'twindragon') { $P1_i -= 3; }
  $points = [ map {$points->[($P1_i-$_)%scalar(@$points)]} 0 .. $#$points ];

  my $Pnum_func = sub {
    my ($i,$Psuffix) = @_;
    $i++;
    if ($type eq 'twindragon') {
      return ($i <= 4 ? "$i$Psuffix" : ($i-4)."${Psuffix}dash");
    }

    my $k = $level - $type_to_level_offset{$type};
    if ($i <  4) { return "$i$Psuffix"; }
    if ($k%4==2 || $k%4==3) {
      if ($i == 4) { return $i."${Psuffix}dash"; }
      $i--;
    }
    if ($i <  8) { return "$i$Psuffix"; }
    if ($k%4==0 || $k%4==3) {
      if ($i == 8) { return $i."${Psuffix}dash"; }
      $i--;
    }
    return "$i$Psuffix";
  };

  $Printwrap = 2;
  printwrap("\\draw[my hull line]\n");
  $Printwrap = 4;
  my $join = '';
  foreach my $p (@$points) {
    my ($x,$y) = @$p;
    ($x,$y) = transform($x,$y);
    printwrap("$join($x,$y)");
    $join = '--';
  }
  printwrap("${join}cycle;\n");

  my $P = $type_to_P{$type};
  my $Psuffix = $type_to_Psuffix{$type};
  $Printwrap = 2;
  foreach my $i (0 .. $#$points) {
    my $p = $points->[$i];
    my ($x,$y) = @$p;
    ($x,$y) = transform($x,$y);
    my $Pnum = $Pnum_func->($i,$Psuffix);
    printwrap(" \\coordinate ($P$Pnum) at ($x,$y);\n");
  }
  $P =~ s/TP/T_P/;
  foreach my $i (0 .. $#$points) {
    my $p = $points->[$i];
    my ($x,$y) = @$p;
    my $Pnum = $Pnum_func->($i,$Psuffix);
    my $GPprefix = ($type eq 'twindragon'
                    ? "BP_"
                    : '');
    print "% GP-Test-Last  $GPprefix$P$Pnum($level) == $x+I*$y\n";
  }
}

print "  % end generated\n";
exit 0;
