#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.010;
use strict;
use warnings;
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $max_level = 10;
if (@ARGV) { $max_level = shift @ARGV; }

print "% generated by tools/$FindBin::Script $max_level\n";
my $path = Math::PlanePath::DragonCurve->new;

my $var = $FindBin::Script;
$var =~ s/\.pl$//;
$var =~ tr/-/_/;

sub path_n_is_right_boundary {
  my ($path, $n, $n_limit) = @_;
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_minus90($dx,$dy);  # to right side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  return (! path_xyxy_is_traversed_either($path, $x1,$y1, $x1s,$y1s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x2,$y2, $x2s,$y2s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x1s,$y1s, $x2s,$y2s, $n_limit));
}
sub path_n_is_left_boundary {
  my ($path, $n, $n_limit) = @_;
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # to side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  ### consider: "$x1,$y1 to $x2,$y2   and $x1s,$y1s to $x2s,$y2s"
  return (! path_xyxy_is_traversed_either($path, $x1,$y1, $x1s,$y1s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x2,$y2, $x2s,$y2s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x1s,$y1s, $x2s,$y2s, $n_limit));
}
sub path_xyxy_is_traversed_either {
  my ($path, $x1,$y1, $x2,$y2, $n_limit) = @_;
  my $n = $path->xyxy_to_n_either($x1,$y1, $x2,$y2);
  return defined $n && $n < $n_limit;
}

my @degree_list = (1, 2, 4);
my @edge_list = ('11','12','22','24','44','44b','44e');

my %counts;
foreach my $k (0 .. $max_level) {
  my ($n_lo, $n_hi) = $path->level_to_n_range($k);

  my %xy_to_degree;
  foreach my $n ($n_lo .. $n_hi-1) {
    my ($x1,$y1) = $path->n_to_xy($n);
    my ($x2,$y2) = $path->n_to_xy($n+1);
    $xy_to_degree{"$x1,$y1"}++;
    $xy_to_degree{"$x2,$y2"}++;
  }
  {
    my %degree_count;
    foreach my $degree (values %xy_to_degree) {
      $degree_count{$degree}++;
    }
    foreach my $degree (@degree_list) {
      push @{$counts{$degree}}, delete $degree_count{$degree} // 0;
    }
    die if %degree_count;
  }

  my %edge_count;
  foreach my $n ($n_lo .. $n_hi-1) {
    my ($x1,$y1) = $path->n_to_xy($n);
    my ($x2,$y2) = $path->n_to_xy($n+1);
    my $degree1 = $xy_to_degree{"$x1,$y1"} // die;
    my $degree2 = $xy_to_degree{"$x2,$y2"} // die;
    if ($degree1 > $degree2) {
      ($degree1,$degree2) = ($degree2,$degree1);
    }
    $edge_count{"$degree1$degree2"}++;

    my $is_boundary = path_n_is_left_boundary($path, $n, $n_hi)
      || path_n_is_right_boundary($path, $n, $n_hi);

    if ($degree1 == 4 && $degree2 == 4) {
      if ($is_boundary) {
        $edge_count{"$degree1${degree2}b"}++;
      } else {
        $edge_count{"$degree1${degree2}e"}++;
      }
    } else {
      $is_boundary
        or die "oops, non-boundary $degree1,$degree2 at k=$k N=$n $x1,$y1 to $x2,$y2";
    }
  }
  foreach my $edge_type (@edge_list) {
    push @{$counts{$edge_type}}, delete $edge_count{$edge_type} // 0;
  }
  if (%edge_count) {
    die "edge_count leftover ",join(' ',keys %edge_count);
  }
}

foreach my $degree (@degree_list) {
  print "% GP-DEFINE  ${var}_DegCount$degree = [";
  my $aref = $counts{$degree};
  print join(',',@$aref);
  print "];\n";
}
foreach my $edge_type (@edge_list) {
  print "% GP-DEFINE  ${var}_EdgeCount$edge_type = [";
  my $aref = $counts{$edge_type};
  print join(',',@$aref);
  print "];\n";
}

print "% end generated\n";
exit 0;
