#!/usr/bin/perl -w

# Copyright 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


# intersecting adjacent convex hull regions

use 5.004;
use strict;
use List::Util 'min';
use Math::Geometry::Planar;
use Math::Geometry::Planar::GPC;

use FindBin;
use lib "$FindBin::Bin/../tools";
use MyDragon '@fVisits_adjacent_list','seg_to_hull6_points';
use MyPlanar;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;

print "  % generated by tools/$FindBin::Script\n";

my $verbose = 0;
my $factor = 3*5;
my $scale = $factor * 6;

my $m_points = seg_to_hull6_points('0,0,0');
$m_points = MyPlanar::points_scale($m_points, $factor);
my $m_gpc = MyPlanar::points_to_gpc($m_points);

my @fVisits_adjacent_points_list
  = map {defined $_
           ? MyPlanar::points_scale(seg_to_hull6_points($_), $factor)
           : undef}
  @fVisits_adjacent_list;
$#fVisits_adjacent_points_list == $#fVisits_adjacent_list or die;

my @fVisits_adjacent_gpc_list
  = map {defined $_ ? MyPlanar::points_to_gpc($_) : undef}
  @fVisits_adjacent_points_list;
$#fVisits_adjacent_gpc_list == $#fVisits_adjacent_list or die;

sub check_gpc_disjoint {
  foreach my $i (0 .. $#_) {
    foreach my $j ($i+1 .. $#_) {
      my $disjoint = Math::Geometry::Planar::GpcClip('INTERSECTION',
                                                     $_[$i],
                                                     $_[$j]);
      MyPlanar::gpc_is_empty($disjoint) || die "oops, not disjoint";
    }
  }
}

my @gpc_overlaps = ($m_gpc,
                    MyPlanar::points_to_gpc([]),
                    MyPlanar::points_to_gpc([]),
                    MyPlanar::points_to_gpc([]),
                    MyPlanar::points_to_gpc([]));
my $max_overlap = $#gpc_overlaps;

foreach my $s (1 .. $#fVisits_adjacent_list) {
  ### $s
  check_gpc_disjoint(@gpc_overlaps);
  my $other_points = $fVisits_adjacent_points_list[$s];
  my $other_gpc    = $fVisits_adjacent_gpc_list[$s];

  ### intersections of $s with existing parts ...
  my @ints = map {
    Math::Geometry::Planar::GpcClip('INTERSECTION', $_, $other_gpc)
    } @gpc_overlaps;

  MyPlanar::gpc_is_empty($ints[$max_overlap])
      or die "oops, overlap at $max_overlap";

  ### n intersection subtract ...
  foreach my $n (0 .. $max_overlap - 1) {
    $gpc_overlaps[$n] = Math::Geometry::Planar::GpcClip
      ('DIFFERENCE', $gpc_overlaps[$n], $ints[$n]);
  }
  ### n intersection add into n+1 ...
  foreach my $n (0 .. $max_overlap - 1) {
    check_gpc_disjoint($gpc_overlaps[$n+1], $ints[$n]);
    $gpc_overlaps[$n+1] = Math::Geometry::Planar::GpcClip
      ('UNION', $gpc_overlaps[$n+1], $ints[$n]);
  }
}


print "  \\begin{scope}[scale=1/$scale]\n";
check_gpc_disjoint(@gpc_overlaps);
foreach my $n (1 .. $max_overlap) {    # no print of 0 others overlap
  my $gpc = $gpc_overlaps[$n];
  my @contours = Math::Geometry::Planar::Gpc2Polygons($gpc);
  next unless @contours;
  my $total_holes = 0;
  $Printwrap = 2;
  printwrap(" \\fill[my $n,even odd rule]");
  $Printwrap += 2;
  foreach my $contour (@contours) {  # Math::Geometry::Planar contour
    ### $contour
    # Second and subsequent polygon of the contour are holes.
    # Think their points already cycle the opposite way as required by pgf
    # default \fill[nonzero rule], but using \fill[even odd rule] works either
    # way.
    my $polygons = $contour->polygons;  # arrayref of points arrayrefs
    $total_holes += scalar(@$polygons)-1;
    my $separator = '';
    foreach my $points (@$polygons) {
      ### $points
      printwrap($separator);
      $separator = "\n";
      printwrap_poly($points);
    }
  }
  printwrap(";\n");
  $total_holes == 0 or die "oops";
}

{
  ### hull sides touch only at points, never in line segments ...

  # $segments[$i] = [ $point1, $point2 ] hull side segment
  my @segments = map { defined $_ ? MyPlanar::poly_to_segments($_) : () }
    $m_points, @fVisits_adjacent_points_list;
  @segments == 12*10 || die;
  foreach my $i (0 .. $#segments) {
    foreach my $j ($i+1 .. $#segments) {
      my $int = MyPlanar::segment_intersection_segment([@{$segments[$i]},
                                                        @{$segments[$j]}]);
      if ($int) {
        MyPlanar::poly_is_point($int) || die;
      }
    }
  }
}

my %extra_points;
{
  my @try_points = @fVisits_adjacent_points_list;
  $try_points[0] = $m_points;

  foreach my $i (0 .. $#try_points) {
    my $i_points = $try_points[$i];
    my @i_segments = MyPlanar::poly_to_segments($i_points);
    my $i_gpc = MyPlanar::points_to_gpc($i_points);

    foreach my $j ($i+1 .. $#try_points) {
      my $j_points = $try_points[$j];
      my $j_gpc = MyPlanar::points_to_gpc($j_points);
      next if MyPlanar::gpc_has_overlap($i_gpc,$j_gpc);  # areas overlap
      my @j_segments = MyPlanar::poly_to_segments($j_points);

      foreach my $is (0 .. $#i_segments) {
        foreach my $js (0 .. $#j_segments) {
          my $int = MyPlanar::segment_intersection_segment
            ([@{$i_segments[$is]},
              @{$j_segments[$js]}]);
          next unless $int;
          MyPlanar::poly_is_point($int) or die "oops, expected point-only";
          my $point = $int->[0];
          next unless gpc_contains_point($m_gpc, $point);
          next unless MyPlanar::point_eq($point, $i_segments[$is]->[0]);
          next unless MyPlanar::point_eq($point, $j_segments[$js]->[0]);

          my $key = join(',',@$point);
          $extra_points{$key}++;
          ### extra point: "$i,$is  $j,$js  $key"
        }
      }
    }
  }
}
while (my ($point_str,$count) = each %extra_points) {
  my $max_n = 0;
  my $point = [split /,/, $point_str];
  foreach my $n (0 .. $max_overlap) {
    if (gpc_contains_point($gpc_overlaps[$n], $point)) {
      $max_n = $n;
    }
  }
  $extra_points{$point_str} += $max_n;
}
### %extra_points

{
  my @point_strs = sort keys %extra_points;
  foreach my $i (0 .. $#point_strs) {
    my $point_str = $point_strs[$i];
    my $count = $extra_points{$point_str};
    print "  \\node (p$i-$count) at ($point_str) [my dot];\n";
  }
}

sub gpc_contains_point {
  my ($gpc, $point) = @_;
  my @contours = Math::Geometry::Planar::Gpc2Polygons($gpc);
  foreach my $contour (@contours) {  # Math::Geometry::Planar contour
    if ($contour->isinside($point)) {
      return 1;
    }
    foreach my $poly (@{$contour->polygons}) {
      if (MyPlanar::poly_has_boundary_point($poly,$point)) {
        return 1;
      }
    }
  }
  return 0;
}

print "  \\end{scope}\n";
print "  % end generated\n";
exit 0;

__END__

my ($region_points,$region_s_aref) = @$region;
my @int_points_list
  = MyPlanar::poly_intersection($region_points, $other_points);

if (@int_points_list) {
  # $region and $other_points intersect in some non-empty area

  my @diff_points_list
    = poly_difference($region_points, $other_points);
  if ($verbose) {
    print "int=",scalar(@int_points_list),
      " diff=",scalar(@diff_points_list),"\n";
  }

  if (@diff_points_list) {
    # $region and $other_points have some intersection and

    foreach my $d_points (@diff_points_list) {
      $d_points // die;
      push @new_regions, [$d_points, $region_s_aref];
    }
    foreach my $i_points (@int_points_list) {
      $i_points // die;
      push @new_regions, [$i_points, [@$region_s_aref, $s]];
    }

    {
      my $union = $diff_points_list[0];
      foreach my $points (@int_points_list, @diff_points_list) {
        my (@union) = MyPlanar::poly_union($union, $points);
        @union == 1 or die;
        $union = $union[0];
      }
      if ($verbose) {
        print "union ",MyDragon::points_str(points_order($union)),"\n";
        print "   cf ",MyDragon::points_str(points_order($region_points)),"\n";
      }
    }
  }

} else {
  push @new_regions, $region;
}
}
@regions = @new_regions;
}

  # my $v = seg_to_hull6_gpc("-1,-1,1");
  # print "v $v\n",points_str(gpc_to_points($v)),"\n";
  # # print points_str(seg_to_hull6_points("0,0,1")),"\n";
  #
  # my $m = seg_to_hull6_gpc("0,0,0");
  # print "m $m\n",points_str(gpc_to_points($m)),"\n";
  # # print points_str(seg_to_hull6_points("0,0,0")),"\n";
  #
  # my $i = Math::Geometry::Planar::GpcClip('INTERSECTION',$m,$v);
  # print "$i\n";
  # print points_str(gpc_to_points($i)),"\n";

  @regions = sort {join(',',@{$a->[1]}) cmp join(',',@{$b->[1]})} @regions;

if ($verbose) {
  foreach my $region (@regions) {
    my ($region_points,$region_s_aref) = @$region;
    printf "%-6s %s\n",
      join(',',@$region_s_aref),
      MyDragon::points_str($region_points);
  }
}

foreach my $i (0 .. $#regions) {
  my ($region_points1,$region_s_aref1) = @{$regions[$i]};
  foreach my $j (0 .. $#regions) {
    next if $i == $j;
    my ($region_points2,$region_s_aref2) = @{$regions[$j]};
    if (! poly_is_disjoint($region_points1,$region_points2)) {
      print "not disjoint\n";
    }
  }
}

print "  \\begin{scope}[scale=1/$scale]\n";
foreach my $region (@regions) {
  my ($region_points,$region_s_aref) = @$region;
  my $count = scalar(@$region_s_aref);
  if ($count != 0) {
    $Printwrap = 2;
    printwrap("\\fill[my $count] ");
    $Printwrap += 2;
    my $join = '';
    foreach my $point (@$region_points) {
      printwrap($join . "($point->[0],$point->[1])");
      $join = '--';
    }
    printwrap("${join}cycle;\n");
  }
}
print "  \\end{scope}\n";

if (0) {
  my $union;
  foreach my $i (0 .. $#regions) {
    my ($region_points,$region_s_aref) = @{$regions[$i]};
    if ($union) {
      my (@union) = MyPlanar::poly_union($union, $region_points);
      @union == 1 or die "oops";
      $union = $union[0];
    } else {
      $union = $region_points;
    }
  }
  print "union ",points_str($union),"\n";
}


sub points_order {
  my ($points) = @_;
  my $min_x;
  my $min_y;
  my $min_i;
  foreach my $i (0 .. $#$points) {
    my $p = $points->[$i];
    if (! defined $min_i
        || $p->[0] < $min_x
        || ($p->[0] == $min_x && $p->[1] < $min_y)) {
      ($min_x,$min_y) = @$p;
      $min_i = $i;
    }
  }
  return [ @{$points}[$min_i..$#$points],
           @{$points}[0 .. $min_i-1] ];
}

sub poly_difference {
  my ($points1, $points2) = @_;
  my $gpc = Math::Geometry::Planar::GpcClip('DIFFERENCE',
                                            MyPlanar::points_to_gpc($points1),
                                            MyPlanar::points_to_gpc($points2));
  if (defined $gpc) {
    return MyPlanar::gpc_to_points_list($gpc);
  }
  return;
}
sub poly_is_disjoint {
  my ($points1, $points2) = @_;
  my @intersection = MyPlanar::poly_intersection($points1,$points2);
  return @intersection == 0;
}

sub seg_to_hull6_gpc {
  my ($seg) = @_;
  return points_to_gpc(seg_to_hull6_points($seg));
}

sub Planar_intersection {
  my $planar = shift;
  my $gpc = $planar->convert2gpc;
  foreach my $p2 (@_) {
    $gpc = Math::Geometry::Planar::GpcClip('INTERSECTION',$gpc,
                                           $p2->convert2gpc);
  }
  return gpc_to_points($gpc);
}
