#!/usr/bin/perl -w

# Copyright 2019 Kevin Hyde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use warnings;
use FindBin;
use List::Util 'min','max','sum';

use lib $FindBin::Bin;
use MyDragon 'bitlength','ffixed_string','ffixed_u','ffixed_du','ffixed_row_w';
use MyPrintwrap ':all';
$|=1;

# uncomment this to run the ### lines
# use Smart::Comments;


my $k_max = 24;
if (@ARGV) { $k_max = shift @ARGV; }

my $k_aggregate = 5;   # num high bits to keep
my $decimals = 8;
use constant option_verbose => 0;
#
# for $k_max, w many blocks to 2^(24+2) etc
# ffixed_row_w(24) ==    27309
# ffixed_row_w(32) ==   699912
# ffixed_row_w(36) ==  3543305
# ffixed_row_w(40) == 17937984


my $zero = 0;

# Should stay accurate within 64 bit int, or 53 bits float and 32 bit int.
# Probably too slow iterating millions of du in bignums.
# use Math::BigInt try => 'GMP';
# use Math::BigFloat;
# $zero = Math::BigInt->new($zero);

print "  % generated by tools/$FindBin::Script $k_max\n";
print "  \\newcommand\\MyKmax{$k_max}\n";

my $n_max = (2 + $zero)**$k_max;

# Return the high bit of $n alone.
sub highbit {
  my ($n) = @_;
  if ($n == 0) { return $n; }
  my $one = 1 + $n*0;
  return $one << bitlength($n) - 1;
}
highbit(5) == 4 or die;
highbit(7) == 4 or die;
highbit(8) == 8 or die;
highbit(9) == 8 or die;


{
  my $str = ffixed_string(num_bits => 4096);

  # Return the number of 1s in the first $n many bits of ffixed.
  sub ffixed_ones {
    my ($n) = @_;
    if ($n <= length($str)) {
      return scalar(substr($str,0,$n) =~ tr/1//);
    } else {
      return undef;
    }
  }
}
ffixed_ones(0) == 0 or die;
ffixed_ones(1) == 1 or die;
ffixed_ones(2) == 2 or die;
ffixed_ones(3) == 2 or die;
ffixed_ones(4) == 3 or die;

{
  my $w_max = ffixed_row_w($k_max-2);
  print "  % w[",$k_max-2,"] = $w_max\n";
}

# my $w_max = ffixed_row_w($k_max-2);
# print "  % w[",$k_max-2,"] = $w_max\n";
# my @du;
# {
#   $du[0] = 4;
#   $du[1] = 4;
#   my $m = 1;
#   while (@du <= $w_max) {
#     # 2h+1 -> 3h+2
#     push @du, 2 * $du[$m++];
#
#     # 2h -> 3h+2
#     push @du, 2 * $du[$m++] - 4;
#
#     # 3h+1 always 4
#     push @du, 4;
#   }
# }
# ### @du
# {
#   my $s = sum(@du);
#   print "  % du[0..$#du] = $s   bitlength=",bitlength($s),"\n";
#   $s == ffixed_u(scalar(@du)) or die;
# }
# foreach my $m (0 .. min(200,$#du)) {
#   ffixed_du($m) == $du[$m] or die;
# }

sub mean {
  my ($ones, $u) = @_;
  $u >= 1 or die;
  if (ref $ones) {
    my $factor = 100_000_000;
    return (($ones * $factor) / $u)->numify / $factor;
  } else {
    return $ones / $u;
  }
}

sub format_number {
  my ($f) = @_;
  $f = sprintf '%.*f', $decimals, $f;
  $f =~ s/0+$//;
  $f =~ s/\.$//;
  return $f;
}

# X coordinate immediately after the column of $n
sub n_to_x {
  my ($n) = @_;
  $n >= 0 or die;
  if ($n==0) { return 0; }
  my $k = bitlength($n);
  my $high = 2**($k-1);
  my $r = $n - $high;
  return format_number($k+$r/$high);
}
# foreach my $n (0 .. 16) {
#   print "n=$n  x = ",n_to_x($n),"\n";
# }
n_to_x(0) == 0 or die;
n_to_x(1) == 1 or die;
n_to_x(2) == 2 or die;
n_to_x(3) == 2.5 or die;
n_to_x(4) == 3 or die;

my $prev_y;
my $n_aggregate = 2**$k_aggregate;
my $x_aggregate = n_to_x($n_aggregate);

{
  # plot lines

  my @x = (0);
  my @y = (mean(ffixed_ones(1), 1));
  $y[0] == 1 or die;
  foreach my $n (1 .. $n_aggregate) {
    my $x1 = n_to_x($n-1);
    my $x2 = n_to_x($n);
    my $y = format_number(mean(ffixed_ones($n), $n));
    push @x, $x1;
    push @y, $y;

    push @x, $x2;
    push @y, $y;
  }

  {
    my $i = 0;
    while ($i <= $#x - 2) {
      if ($y[$i] eq $y[$i+1] && $y[$i+1] eq $y[$i+2]) {
        ### delete co-linear: "i=$i y=$y[$i]"
        splice @x, $i+1, 1;
        splice @y, $i+1, 1;
      } else {
        $i++;
      }
    }
  }


  $Printwrap = 2;
  printwrap("\\draw[line cap=butt] ");
  $Printwrap += 2;
  my $join = '';
  foreach my $i (0 .. $#x) {
    printwrap("$join($x[$i],$y[$i])");
    $join = '--';
  }
  $prev_y = $y[-1];
}

use constant ffixed_u_uB_first => (0,0);
sub ffixed_u_uB_next {
  my ($u, $uB) = @_;
  ### ffixed_u_uB_next: "$u  $uB"
  if ($u == 0) { return (4,0); }
  my $mask = $u ^ ($u-1);
  my $high = $mask ^ ($mask>>1);
  ### $mask
  ### $high
  return (($u | ($uB & $mask)) + 4,
          $high | ($uB & ~$mask));
}
{
  my ($u,$uB) = ffixed_u_uB_first();
  foreach my $m (0 .. 200) {
    $u == ffixed_u($m) or die "iterate m=$m  $u $uB";
    ($u,$uB) = ffixed_u_uB_next($u,$uB);
  }
}

{
  my $m = $zero;
  my ($u,$uB) = ffixed_u_uB_first();
  $u += $zero;
  $uB += $zero;
  my $upto_u = $zero;
  my @queue_len;
  my @queue_value;
  sub take_some {
    my ($max_len) = @_;
    unless (@queue_len) {
      if ($m < 500) {
        ($upto_u == $u && $u == ffixed_u($m)) or die "oops, m=$m u=$u";
      }
      (my $u_next, $uB) = ffixed_u_uB_next($u,$uB);
      my $du = $u_next - $u;
      if (option_verbose()) { print "take m=$m  du=$du\n"; }
      if (ref $zero) { ref $du or die "du not a BigInt at m=$m"; }
      if ($m & 1) {  # after m odd is zeros
        push @queue_len,   $du;
        push @queue_value, 0;
      } else {
        ### queue ones ...
        push @queue_len,   1, 1, $du-2;
        push @queue_value, 1, 0, 1;
      }
      $m++;
      $u = $u_next;
    }
    my $value = $queue_value[-1];
    my $len = min($queue_len[-1], $max_len);
    $queue_len[-1] -= $len;
    $queue_len[-1] >= 0 or die;
    if ($queue_len[-1] == 0) {
      ### pop empty length of value: $value
      pop @queue_len;
      pop @queue_value;
    }
    $upto_u += $len;
    ### to: "len=$len is u=$u"
    return ($len, $upto_u, $value);
  }
}

my $bucket_lo = 1 + $zero;
my $bucket_hi = 1 + $zero;
sub next_bucket {
  ### next_bucket() ...
  $bucket_lo = $bucket_hi + 1;
  $bucket_hi = $bucket_lo;
  my $mask = highbit($bucket_hi-1) - 1;
  $mask >>= ($k_aggregate - 1);
  $bucket_hi += $mask;
  $bucket_hi >= $bucket_lo or die;
}
# foreach (1 .. 64) {
#   print "bucket $bucket_lo to $bucket_hi  length ",$bucket_hi-$bucket_lo+1,"\n";
#   next_bucket();
# }

sub check_ones {
  my ($ones, $u, $name) = @_;
  if (defined(my $want_ones = ffixed_ones($u))) {
    $ones == $want_ones
      or die "$name oops want_ones $want_ones at u=$u, got $ones";
  }
}

# generate aggregated
my @bucket_lo;
my @bucket_hi;
my @bucket_mean_min;
my @bucket_mean_max;

my $ones = 0;
my $prev_u = 0;
while ($bucket_hi <= $n_max) {
  ### bucket: "$bucket_lo to $bucket_hi"
  if (option_verbose()) {
    print "bucket: $bucket_lo to $bucket_hi  count $#bucket_lo  ones $ones\n";
  }

  my $bucket_mean_min;
  my $bucket_mean_max;

  for (;;) {
    my $this_max_len = $bucket_hi - $prev_u;
    ### at: "prev_u=$prev_u in bucket $bucket_lo to $bucket_hi  this_max_len $this_max_len"

    my ($len,$u,$value) = take_some($this_max_len);
    ### some: "len=$len u=$u value=$value  into bucket $bucket_lo $bucket_hi"
    $len >= 0 or die;
    $u <= $bucket_hi or die "oops, u=$u bucket_hi=$bucket_hi";
    if ($len==0) {
      ### end of bucket ...
      last;
    }
    {
      my $start_ones = $ones + 1*$value;
      my $start_u = $u - $len + 1;
      check_ones ($start_ones, $start_u, 'start');
      my $mean_start = mean($start_ones, $start_u);
      $bucket_mean_min //= $mean_start;
      $bucket_mean_max //= $mean_start;
      $bucket_mean_min = min($bucket_mean_min, $mean_start);
      $bucket_mean_max = max($bucket_mean_max, $mean_start);
    }
    $ones += $len*$value;
    {
      check_ones($ones, $u, 'end');
      my $mean_end = mean($ones, $u);
      $bucket_mean_min = min($bucket_mean_min, $mean_end);
      $bucket_mean_max = max($bucket_mean_max, $mean_end);
    }
    $prev_u = $u;
  }

  $prev_u == $bucket_hi or die;

  if ($bucket_lo > $n_aggregate) {
    ### new bucket ...
    if (@bucket_lo == 0) {
      $bucket_lo == $n_aggregate+1
        or die "oops bucket_lo $bucket_lo vs n_aggregate $n_aggregate";
      $bucket_hi > $bucket_lo or die;
    }
    push @bucket_lo, $bucket_lo;
    push @bucket_hi, $bucket_hi;
    push @bucket_mean_min, format_number($bucket_mean_min);
    push @bucket_mean_max, format_number($bucket_mean_max);
  }
  next_bucket();
}

### bucket_lo: map {"$bucket_lo[$_] to $bucket_hi[$_]  mean $bucket_mean_min[$_] to $bucket_mean_max[$_]"} 0 .. $#bucket_lo

# plot aggregated

# end the plot line down to the aggregate if necessary
if ($prev_y < $bucket_mean_min[0]) {
  printwrap("--($x_aggregate,$bucket_mean_min[0]);\n");
} elsif ($prev_y > $bucket_mean_max[0]) {
  printwrap("--($x_aggregate,$bucket_mean_max[0]);\n");
} else {
  printwrap(";\n");
}

print "  % fill ",scalar(@bucket_lo)," buckets\n";
{
  $Printwrap = 2;
  printwrap("\\fill [shift={(line-width-lo)}] ");
  $Printwrap += 2;
  my $join = '';
  foreach my $i (0 .. $#bucket_lo) {
    my $x1 = n_to_x($bucket_lo[$i]-1);
    my $x2 = n_to_x($bucket_hi[$i]);
    my $y = $bucket_mean_min[$i];
    printwrap("$join($x1,$y)");
    printwrap("--($x2,$y)");
    $join = '--';
  }
  foreach my $i (reverse 0 .. $#bucket_lo) {
    my $x1 = n_to_x($bucket_hi[$i]);
    my $x2 = n_to_x($bucket_lo[$i]-1);
    my $y = $bucket_mean_max[$i];
    printwrap("--($x1,$y)");
    printwrap("--($x2,$y)");
  }
  printwrap(" --cycle;\n");
}

$bucket_hi[-1] == $n_max or die;
my $x_max = n_to_x($n_max);
print "  \\coordinate (x-max) at ($x_max, 0);\n";

# # X ticks
# $Printwrap = 2;
# printwrap("\\draw");
# $Printwrap += 2;
# foreach my $k ($level_lo+1 .. $level_hi) {
#   my $n = 2**$k;
#   my $x = n_to_x($n);
#   my $semi = ($k==$level_hi ? ";\n" : "");
#   printwrap(" ($x,1/2) -- +(x-tick)$semi");
# }
#
# # X numbers
# print "  \\begin{scope}[font=\\scriptsize]\n";
# foreach my $k ($level_lo .. $level_hi-1) {
#   my $n = 2**$k;
#   my $x = n_to_x($n);
#   print "  \\node at ($x,1/2) [below right,my x number] {\$$n\$};\n";
# }
# print "  \\end{scope}\n";

print "  % end generated\n";
exit 0;
