#!/usr/bin/perl -w

# Copyright 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use Math::PlanePath::ComplexPlus 117; # v.117 for level_to_n_range()
use Tie::IxHash;

use FindBin;
use lib $FindBin::Bin;
use MyDragon '@dir4_to_dx','@dir4_to_dy',
  'xy_rotate_plus90','mul','div';
use MyLineSegments;
use MyPrintwrap ':all';


my $level = 5;
if (@ARGV) { $level = shift @ARGV; }

print "  % generated by tools/$FindBin::Script\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

my $path = Math::PlanePath::ComplexPlus->new;
my ($n_lo, $n_hi) = $path->level_to_n_range($level);

sub is_visited {
  my ($x,$y) = @_;
  my $n = $path->xy_to_n($x,$y);
  return (defined $n && $n <= $n_hi);
}


{
  # complex base unit squares

  my $segments = MyLineSegments->new;
  foreach my $n ($n_lo .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    foreach my $dir4 (0 .. 3) {
      my $dx = $dir4_to_dx[$dir4];
      my $dy = $dir4_to_dy[$dir4];
      my $x2 = $x + $dx;
      my $y2 = $y + $dy;
      $segments->add(2*$x,2*$y,  2*$x2,2*$y2);
      ($x,$y) = ($x2,$y2);
    }
  }
  $Printwrap = 2;
  printwrap("\\draw[my square lines] ");
  $Printwrap += 2;
  $segments->printwrap;
  printwrap(";\n");
}

{
  # twindragon segments and points

  my $transform = sub {
    my ($x,$y) = @_;
    ($x,$y) = div($x,$y);
    return ($x/2, $y/2);
  };
  my $untransform = \&mul;

  my $segments = MyLineSegments->new;
  my %points;
  tie %points, 'Tie::IxHash';
  foreach my $n ($n_lo .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    foreach my $dir4 (0 .. 3) {
      my $dx = $dir4_to_dx[$dir4];
      my $dy = $dir4_to_dy[$dir4];
      my ($lx,$ly) = xy_rotate_plus90($dx,$dy);
      my ($x2,$y2) = ($x + $dx, $y + $dy);
      my ($sx,$sy) = (2*$x+$dx, 2*$y+$dy);
      my ($tx,$ty) = (2*$x2+$lx, 2*$y2+$ly);

      $segments->add($untransform->($sx,$sy),
                     $untransform->($tx,$ty));
      $points{"$sx,$sy"}++;
      ($x,$y) = ($x2,$y2);
    }
  }
  $Printwrap = 2;
  printwrap("\\draw ");
  $Printwrap += 2;
  $segments->printwrap (transform => $transform);
  printwrap(";\n");

  $Printwrap = 2;
  printwrap("\\path foreach \\p in {");
  $Printwrap += 2;
  foreach my $xy (keys %points) {
    printwrap("($xy),");
  }
  printwrap("}");
  printwrap("{ \\p node[my dot] };\n");

  my $TP = scalar keys %points;
  my $TS = scalar grep {$points{$_}==1} keys %points;
  my $TD = scalar grep {$points{$_}==2} keys %points;
  print "% GP-Test  TP($level) == $TP\n";
  print "% GP-Test  TS($level) == $TS\n";
  print "% GP-Test  TD($level) == $TD\n";
}

{
  # graph

  my $segments = MyLineSegments->new;
  my @points;
  foreach my $n ($n_lo .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    my ($sx,$sy) = (2*$x+1,2*$y+1);
    push @points, "$sx,$sy";
    foreach my $dir4 (0 .. 3) {
      my $dx = $dir4_to_dx[$dir4];
      my $dy = $dir4_to_dy[$dir4];
      my ($lx,$ly) = xy_rotate_plus90($dx,$dy);
      my $x2 = $x + $dx;
      my $y2 = $y + $dy;
      next unless is_visited($x2,$y2);

      #      B     B
      #   *-----*-----*      at least one of the B
      #   | x,y |     |      squares not visited so
      #   *-----*-----*      boundary
      #      B     B

      if (! is_visited($x+$lx,$y+$ly)
          || ! is_visited($x-$lx,$y-$ly)
          || ! is_visited($x2+$lx,$y2+$ly)
          || ! is_visited($x2-$lx,$y2-$ly)) {
        $segments->add($sx,$sy, 2*$x2+1, 2*$y2+1);
      }
    }
  }

  $Printwrap = 2;
  printwrap("\\begin{scope}[shift=(pos-graph)]\n");
  $Printwrap += 2;
  printwrap("\\draw ");
  $Printwrap += 2;
  $segments->printwrap;
  printwrap(";\n");

  $Printwrap = 4;
  printwrap("\\path foreach \\p in {");
  $Printwrap += 2;
  foreach my $xy (@points) {
    printwrap("($xy),");
  }
  printwrap("}");
  printwrap(" { \\p node[my square dot] };\n");

  my $num_vertices = scalar(@points);
  my @s = $segments->segments;
  my $num_edges = scalar(@s) / 2;
  print "% GP-Test  EightB($level) == $num_vertices\n";
  print "% GP-Test  TDB($level) == $num_edges\n";
  print "  \\end{scope}\n";
}

print "  % end generated\n";
exit 0;
